#include "defs.h"

#define MAXHL			64
#define MAXFL			(2*MAXHL+1)
#define MINFILTER		1e-4
#define PI			3.141592653589793

#define C0MAX		(100.0)
#define C0MIN		(0.0)
#define C1MAX		(100.0)
#define C1MIN		(-100.0)
#define C2MAX		(100.0)
#define C2MIN		(-100.0)
 
#define G0MAX		(C0MAX - C0MIN)
#define G0MIN		(0)
#define G1MAX		((C1MAX - C1MIN)/4)
#define G1MIN		(0)
#define G2MAX		((C2MAX - C2MIN)/4)
#define G2MIN		(0)

#define WAVELET_THRESHOLD	300

static double L1Norm   (float *f1, float *f2, int length);
static void   normalize(float *X, float *Y, int slen);
static void   convolve (float *X, float *H, float *Y, int slen, int hl);
static void   convolve2(float *X, float *H, float *Y, int slen, int hl);

static int  C0toBIN(float x);
static int  C1toBIN(float x);
static int  C2toBIN(float x);

static int  G0toBIN(float value);
static int  G1toBIN(float value);
static int  G2toBIN(float value);

static int  B0toBIN(float angle);
static int  B1toBIN(float angle);
static int  B2toBIN(float angle);

static int write_int  (FILE *fp, int *value, int ncount);
static int read_int   (FILE *fp, int *value, int ncount);
static int write_hist (FILE *fp, float *histogram, int points, int size);
static int read_hist  (FILE *fp, float *histogram, int points, int size);
static int color_to_texture(struct float_img *data_img, struct float_img *texture);
static int color_to_edgemap(struct float_img *data_img, struct float_img *edgemap);
static int color_to_gradient (struct float_img *data_img, 
                              struct float_img *gradr_img, 
		                      struct float_img *gradb_img);
static int color_to_laplacian(struct float_img *data_img, struct float_img *lapl_img);
static int color_to_wavelet(struct float_img *data_img, struct float_img *wavelet);
static void wavelet2(
					float	**data,
					float	**output,
					int		width,
					int		height,
				    float 	*filt,  	      /*  QMF filter coefficients  */
					int 	flen); 		      /*  QMF filter length  */
static void wavelet(float	*data,
					float	*output,
					int		dlen,
					float 	*filt,  	      /*  QMF filter coefficients  */
					int 	flen); 		      /*  QMF filter length  */
static void anqmf2(
			    float	**data,
				float	**output,
				int		width,
				int		height,
				float 	*filt,  	      /*  QMF filter coefficients  */
				int 	flen); 		      /*  QMF filter length  	   */
static void anqmf(
				float	*data,
				float	*output,
				int		dlen,
				float 	*filt,  	      /*  QMF filter coefficients  */
				int 	flen); 		      /*  QMF filter length  */

static float  weight[9]  = {1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0};

static int   HL = MAXHL;
static int   FL = MAXFL;
static float normal[MAXFL];
#define MASKSIZE	3

static 	float mask[3][3] = {0, -1, 0, -1, 4, -1, 0, -1, 0};
static 	float filter[5][5] = {	0.0029,	0.0131,	0.0215,	0.0131,	0.0029,
				0.0131,	0.0585,	0.0965,	0.0585,	0.0131,
				0.0215,	0.0965,	0.1592,	0.0965,	0.0215,
				0.0131,	0.0585,	0.0965,	0.0585,	0.0131,
				0.0029,	0.0131,	0.0215,	0.0131,	0.0029};
static float haar[2] = {1, 1};


void build_feature(struct float_img *fimg, struct feature *feature)
{
  int i1, j1, l, pts, peakx, peaky;
  float peak;
  struct float_img edgemap, texture, wavelet;

  feature->col = fimg->col;
  feature->row = fimg->row;

  color_to_texture(fimg, &texture); 
  color_to_edgemap(fimg, &edgemap); 

  for (l=0; l<HIST0_BINS; l++) feature->hist0[l] = 0;
  for (l=0; l<HIST1_BINS; l++) feature->hist1[l] = 0;
  for (l=0; l<HIST2_BINS; l++) feature->hist2[l] = 0;
  for (l=0; l<HIST3_BINS; l++) feature->hist3[l] = 0;
  for (l=0; l<HIST4_BINS; l++) feature->hist4[l] = 0;
  for (l=0; l<HIST5_BINS; l++) feature->hist5[l] = 0;
  for (l=0; l<HIST6_BINS; l++) feature->hist6[l] = 0;
  for (l=0; l<HIST7_BINS; l++) feature->hist7[l] = 0;
  for (l=0; l<HIST8_BINS; l++) feature->hist8[l] = 0;

  feature->pts = 0;
  feature->status = 0;
  for(i1=0; i1<fimg->row; i1++)
  for(j1=0; j1<fimg->col; j1++) {

    feature->pts ++;
    feature->hist0[C0toBIN(  fimg->color[0][i1][j1])] ++;
    feature->hist1[C1toBIN(  fimg->color[1][i1][j1])] ++;
    feature->hist2[C2toBIN(  fimg->color[2][i1][j1])] ++;
    feature->hist3[G0toBIN(texture.color[0][i1][j1])] ++;
    feature->hist4[G1toBIN(texture.color[1][i1][j1])] ++;
    feature->hist5[G2toBIN(texture.color[2][i1][j1])] ++;
    feature->hist6[B0toBIN(edgemap.color[0][i1][j1])] ++;
    feature->hist7[B1toBIN(edgemap.color[1][i1][j1])] ++;
    feature->hist8[B2toBIN(edgemap.color[2][i1][j1])] ++;

  }

#ifdef WAVELET
  color_to_wavelet(fimg, &wavelet);
  for (i1=0; i1<wavelet.row; i1++) 
  for (j1=0; j1<wavelet.col; j1++) {
	wavelet.color[0][i1][j1] = fabs(wavelet.color[0][i1][j1]);
  }

  for (l=0; l<WAVELET_BINS; l++) {
	peak = -1;
	peakx = 0;
	peaky = 0;
	for (i1=0; i1<wavelet.row; i1++) 
	for (j1=0; j1<wavelet.col; j1++) {
		if (peak < wavelet.color[0][i1][j1]) {
			peak = wavelet.color[0][i1][j1];
			peaky = i1;
			peakx = j1;
		}
	}
    wavelet.color[0][peaky][peakx] = -1;
	feature->peakx[l] = peakx;
	feature->peaky[l] = peaky;
  }
  for (l=0; l<3; l++) free_img((void **) wavelet.color[l]);
  free((void *) wavelet.color );
#endif

  pts = feature->pts;


  normalize(feature->hist0, feature->hist0, HIST0_BINS);

  normalize(feature->hist1, feature->hist1, HIST1_BINS);

  normalize(feature->hist2, feature->hist2, HIST2_BINS);

  normalize(feature->hist3, feature->hist3, HIST3_BINS);

  normalize(feature->hist4, feature->hist4, HIST4_BINS);

  normalize(feature->hist5, feature->hist5, HIST5_BINS);

  normalize(feature->hist6, feature->hist6, HIST6_BINS);

  normalize(feature->hist7, feature->hist7, HIST7_BINS);

  normalize(feature->hist8, feature->hist8, HIST8_BINS);


  for (l=0; l<3; l++) free_img((void **) texture.color[l]);
  free((void *) texture.color );

  for (l=0; l<3; l++) free_img((void **) edgemap.color[l]);
  free((void *) edgemap.color );

  for (l=0; l<3; l++) free_img((void **) fimg->color[l]);
  free((void *) fimg->color );


}  


void naturalize(struct feature *feature) 
{
  float  f1[256], f2[256];

  convolve(feature->hist0, normal, f1, HIST0_BINS, HL);
  convolve(f1, normal, f2, HIST0_BINS, HL);
  normalize(f1, feature->hist0, HIST0_BINS);

  convolve(feature->hist1, normal, f1, HIST1_BINS, HL);
  convolve(f1, normal, f2, HIST1_BINS, HL);
  normalize(f1, feature->hist1, HIST1_BINS);

  convolve(feature->hist2, normal, f1, HIST2_BINS, HL);
  convolve(f1, normal, f2, HIST2_BINS, HL);
  normalize(f1, feature->hist2, HIST2_BINS);

  convolve(feature->hist3, normal, f1, HIST3_BINS, HL);
  convolve(f1, normal, f2, HIST3_BINS, HL);
  normalize(f1, feature->hist3, HIST3_BINS);

  convolve(feature->hist4, normal, f1, HIST4_BINS, HL);
  convolve(f1, normal, f2, HIST4_BINS, HL);
  normalize(f1, feature->hist4, HIST4_BINS);

  convolve(feature->hist5, normal, f1, HIST5_BINS, HL);
  convolve(f1, normal, f2, HIST5_BINS, HL);
  normalize(f1, feature->hist5, HIST5_BINS);
/*
  convolve2(feature->hist6, normal, f1, HIST6_BINS, HL);
  normalize(f1, feature->hist6, HIST6_BINS);

  convolve2(feature->hist7, normal, f1, HIST7_BINS, HL);
  normalize(f1, feature->hist7, HIST7_BINS);

  convolve2(feature->hist8, normal, f1, HIST8_BINS, HL);
  normalize(f1, feature->hist8, HIST8_BINS);
*/
}



double feature_distance(struct feature *dfeature, struct feature *tfeature) 
{
  double L0, L1, L2, L3, L4, L5, L6, L7, L8, L9, L;
  float  f1[256];
  int    i, j;

  if (weight[0] > 0) {
    if (tfeature->status < 1) {
      convolve(tfeature->hist0, normal, f1, HIST0_BINS, HL);
      normalize(f1, tfeature->hist0, HIST0_BINS);
      tfeature->status = 1;
    }
    if (dfeature->status < 1) {
      convolve(dfeature->hist0, normal, f1, HIST0_BINS, HL);
      normalize(f1, dfeature->hist0, HIST0_BINS);
      dfeature->status = 1;
    }
  }
  if (weight[1] > 0) {
    if (tfeature->status < 2) {
      convolve(tfeature->hist1, normal, f1, HIST1_BINS, HL);
      normalize(f1, tfeature->hist1, HIST1_BINS);
      tfeature->status = 2;
    }
    if (dfeature->status < 2) {
      convolve(dfeature->hist1, normal, f1, HIST1_BINS, HL);
      normalize(f1, dfeature->hist1, HIST1_BINS);
      dfeature->status = 2;
    }
  }
  if (weight[2] > 0) {
    if (tfeature->status < 3) {
      convolve(tfeature->hist2, normal, f1, HIST2_BINS, HL);
      normalize(f1, tfeature->hist2, HIST2_BINS);
      tfeature->status = 3;
    }
    if (dfeature->status < 3) {
      convolve(dfeature->hist2, normal, f1, HIST2_BINS, HL);
      normalize(f1, dfeature->hist2, HIST2_BINS);
      dfeature->status = 3;
    }
  }

  L0 = L1 = L2 = L3 = L4 = L5 = L6 = L7 = L8 = L9 = 0.0;

  if (weight[0] > 0)
    L0 = L1Norm(dfeature->hist0, tfeature->hist0, HIST0_BINS) * weight[0];
  if (weight[1] > 0)
    L1 = L1Norm(dfeature->hist1, tfeature->hist1, HIST1_BINS) * weight[1];
  if (weight[2] > 0)
    L2 = L1Norm(dfeature->hist2, tfeature->hist2, HIST2_BINS) * weight[2];
  if (weight[3] > 0)
    L3 = L1Norm(dfeature->hist3, tfeature->hist3, HIST3_BINS) * weight[3];
  if (weight[4] > 0)
    L4 = L1Norm(dfeature->hist4, tfeature->hist4, HIST4_BINS) * weight[4];
  if (weight[5] > 0)
    L5 = L1Norm(dfeature->hist5, tfeature->hist5, HIST5_BINS) * weight[5];
  if (weight[6] > 0)
    L6 = L1Norm(dfeature->hist6, tfeature->hist6, HIST6_BINS) * weight[6];
  if (weight[7] > 0)
    L7 = L1Norm(dfeature->hist7, tfeature->hist7, HIST7_BINS) * weight[7];
  if (weight[8] > 0)
    L8 = L1Norm(dfeature->hist8, tfeature->hist8, HIST8_BINS) * weight[8];

#ifdef WAVELET
  L9 = WAVELET_BINS;
  for (i=0; i<WAVELET_BINS; i++) 
    for (j=0; j<WAVELET_BINS; j++) 
		if (tfeature->peakx[i] == dfeature->peakx[j] && tfeature->peaky[i] == dfeature->peaky[j]) {
			L9 -= 1;
			break;
		}
  L9 /= WAVELET_BINS;
  L9 *= 2.0;
#endif

  L  = L0 + L1 + L2 + L3 + L4 + L5 + L6 + L7 + L8 + L9;

  return(L);

}


static double L1Norm(float *f1, float *f2, int length)
{
  int    i;
  double L;
  
  L = 0.0;
  for (i=0; i<length; i++) {
    L += fabs(f1[i] - f2[i]);
  }

  return(L);
}

static void normalize(float *X, float *Y, int slen)
{
  int   l;
  float sum;

  for (sum=0, l=0; l<slen; l++) sum  += X[l];
  for (       l=0; l<slen; l++) Y[l] = X[l] / sum;

}

static void convolve(float *X, float *H, float *Y, int slen, int hl)
{
  int i, j;
  float sum;

  for (i=0; i<slen; i++) Y[i] = 0;
  for (i=0; i<slen; i++) {
    if (X[i] > 0) {

      for (j=max(0, i-hl), sum=0; j<=min(slen-1, i+hl); j++) sum += H[i+hl-j];
      for (j=max(0, i-hl);        j<=min(slen-1, i+hl); j++) {
        Y[j] += X[i]*H[i+hl-j] / sum;

      }
    }
  }

}

static void convolve2(float *X, float *H, float *Y, int slen, int hl)
{
  int i, j, k;

  for (i=0; i<slen; i++) Y[i] = 0;
  for (i=0; i<slen-1; i++) {
    if (X[i] > 0) {

      for (j=i-hl; j<=i+hl; j++) {
        k = (j + slen - 1) % (slen - 1);
        Y[k] += X[i]*H[i+hl-j];

      }
    }
  }
  Y[slen-1] = X[slen-1];

}

void setWeight(float *newweight)
{
    int i;

    for (i=0; i<9; i++) weight[i] = newweight[i];

}

void setFilter(float sigma)
{
    float sum, T2;
    float f0[MAXFL];
    int i;

    sum = 0;
    T2 = (float) sqrt(2 * PI) * sigma;
    for (i=0; i<MAXFL; i++) {
      if (sigma <= 0) {
        if (i == MAXHL) f0[i] = 1.0;
        else            f0[i] = 0;
      } else {
	f0[i] = (float) exp(-(i-MAXHL)*(i-MAXHL)/(2*sigma*sigma))/T2;
      }
    }

    for (i=0, FL=0; i<MAXFL; i++) 
      if (f0[i] > MINFILTER) normal[FL++] = f0[i];
    for (i=0, sum=0; i<FL; i++) sum += normal[i];
    for (i=0;        i<FL; i++) normal[i] /= sum;
    HL = FL / 2;
/*
    fprintf(stdout, "FILTER %d:", FL);
    for (i=0;        i<FL; i++) fprintf(stdout, " %g", normal[i]);
    fprintf(stdout, "\n");
*/

}

static int C0toBIN(float x)
{
  int index;

  index = (int) floor((x - C0MIN) / (C0MAX - C0MIN) * HIST0_BINS);
  index = min(HIST0_BINS-1, max(0, index));
  return(index);

}

static int C1toBIN(float x)
{
  int index;

  index = (int) floor((x - C1MIN) / (C1MAX - C1MIN) * HIST1_BINS);
  index = min(HIST1_BINS-1, max(0, index));
  return(index);

}

static int C2toBIN(float x)
{
  int index;

  index = (int) floor((x - C2MIN) / (C2MAX - C2MIN) * HIST2_BINS);
  index = min(HIST2_BINS-1, max(0, index));
  return(index);

}

static int G0toBIN(float x)
{
  int index;

  index = (int) floor((x - G0MIN) / (G0MAX - G0MIN) * HIST3_BINS);
  index = min(HIST3_BINS-1, max(0, index));

  return(index);

}

static int G1toBIN(float x)
{
  int index;

  index = (int) floor((x - G1MIN) / (G1MAX - G1MIN) * HIST4_BINS);
  index = min(HIST4_BINS-1, max(0, index));

  return(index);

}

static int G2toBIN(float x)
{
  int index;

  index = (int) floor((x - G2MIN) / (G2MAX - G2MIN) * HIST5_BINS);
  index = min(HIST5_BINS-1, max(0, index));

  return(index);

}

static int B0toBIN(float angle)
{
  int index;

  index = (int) floor((angle + PI) * (HIST6_BINS-1) / (2 * PI) + 0.5);
  if (index <  0)            index = HIST6_BINS-1;
  else                       index = index % (HIST6_BINS-1);

  return(index);

}

static int B1toBIN(float angle)
{
  int index;

  index = (int) floor((angle + PI) * (HIST7_BINS-1) / (2 * PI) + 0.5);
  if (index <  0)            index = HIST7_BINS-1;
  else                       index = index % (HIST7_BINS-1);

  return(index);

}

static int B2toBIN(float angle)
{
  int index;

  index = (int) floor((angle + PI) * (HIST8_BINS-1) / (2 * PI) + 0.5);
  if (index <  0)            index = HIST8_BINS-1;
  else                       index = index % (HIST8_BINS-1);

  return(index);

}

static int write_hist(FILE *fp, float *histogram, int points, int size)
{
  int		i, value;
  unsigned char buffer, start, stop, fileptr, filebuf[256];

  for (start=0;     start<size; start++) if (histogram[start] != 0) break;
  for (stop=size-1; stop>0;     stop--)  if (histogram[stop]  != 0) break;

  for (i=start, fileptr=0; i<=stop && i<size; i++) {
    value = (int) floor(histogram[i] * points + 0.5);

    while (value >= 128) {
      buffer = (unsigned char) (value % 128) | 0x80; 
      filebuf[fileptr++] = buffer;
      value /= 128;
    } 
    buffer = (unsigned char) value;
    filebuf[fileptr++] = buffer;

  }
  fwrite(&start,   sizeof(unsigned char), 1, fp);
  fwrite(&fileptr, sizeof(unsigned char), 1, fp);
  fwrite(filebuf,  sizeof(unsigned char), fileptr, fp);

  return(SUCCESSED);
}

static int read_hist(FILE *fp, float *histogram, int points, int size)
{
  int		i, value, multiple;
  unsigned char buffer, start, length, fileptr, filebuf[256];

  fread(&start,  sizeof(unsigned char), 1, fp);
  fread(&length, sizeof(unsigned char), 1, fp);
  fread(filebuf, sizeof(unsigned char), length, fp);
 
  for (i=0; i<size; i++) histogram[i] = 0;
  for (i=start, fileptr=0; fileptr<length && i<size; i++) {
    buffer = filebuf[fileptr++];
    value = 0;
    multiple = 1;
    while ((buffer & 0x80) == 0x80) {
      value += ((int) (buffer & 0x7f)) * multiple; 
      buffer = filebuf[fileptr++];
      multiple *= 128;
    }
    value += ((int) (buffer & 0x7f)) * multiple; 
    histogram[i] = ((float) value) / points;
  }

  return(SUCCESSED);
}

static int write_int(FILE *fp, int *value, int ncount)
{
  int i, temp;
  unsigned char buffer;

  for (i=0; i<ncount; i++) {
    temp = value[i];
    while (temp >= 128) {
      buffer = (unsigned char) (temp % 128) | 0x80; 
      fwrite(&buffer, sizeof(unsigned char), 1, fp);
      temp /= 128;
	} 
    buffer = (unsigned char) temp;
    fwrite(&buffer, sizeof(unsigned char), 1, fp);
  }


  /* Big Endian */
/*
  long ldummy=0x01020304;
  unsigned char *cdummy;
  int  i, j;
  cdummy = (unsigned char *) &ldummy;
  if (cdummy[0] == 0x01) {
    for (j=0; j<ncount; j++) {
      for (i=0; i<4; i++) {
        cdummy = (unsigned char *) &value[j];
        fwrite(&cdummy[i], sizeof(unsigned char), 1, fp);
      }
    }
  } else {
    for (j=0; j<ncount; j++) {
      for (i=3; i>=0; i--) {
        cdummy = (unsigned char *) &value[j];
        fwrite(&cdummy[i], sizeof(unsigned char), 1, fp);
      }
    }
  }
*/
  return(SUCCESSED);
}

static int read_int(FILE *fp, int *value, int ncount)
{
  int i, temp, multiple;
  unsigned char buffer;

  for (i=0; i<ncount; i++) {
    temp = 0;
    multiple = 1;
    fread(&buffer, sizeof(unsigned char), 1, fp);
    while ((buffer & 0x80) == 0x80) {
      temp += ((int) (buffer & 0x7f)) * multiple; 
      fread(&buffer, sizeof(unsigned char), 1, fp);
      multiple *= 128;
	}
    temp += ((int) (buffer & 0x7f)) * multiple; 
    value[i] = temp;
  }

    /* Big Endian */
/*
  long ldummy=0x01020304;
  unsigned char *cdummy;
  int  i, j;
  cdummy = (unsigned char *) &ldummy;
  if (cdummy[0] == 0x01) {
    for (j=0; j<ncount; j++) {
      for (i=0; i<4; i++) {
        cdummy = (unsigned char *) &value[j];
        fread(&cdummy[i], sizeof(unsigned char), 1, fp);
      }
    }
  } else {
    for (j=0; j<ncount; j++) {
      for (i=3; i>=0; i--) {
        cdummy = (unsigned char *) &value[j];
        fread(&cdummy[i], sizeof(unsigned char), 1, fp);
      }
    }
  }
*/
  return(SUCCESSED);
}

int write_feature(struct feature *feature, FILE *fp, int *lstart, int *len)
{
  int i;

  *lstart = ftell(fp);

  if (feature->status != 0) feature->pts = max(1000, feature->pts);
  write_int(fp, &(feature->pts),    1);
  write_int(fp, &(feature->status), 1);

  if (write_hist(fp, feature->hist0, feature->pts, HIST0_BINS) == FAILED) {
    fprintf(stderr, "Error writing histogram L of %dpts\n", feature->pts);
    return(FAILED);
  }
  if (write_hist(fp, feature->hist1, feature->pts, HIST1_BINS) == FAILED) {
    fprintf(stderr, "Error writing histogram A of %dpts\n", feature->pts);
    return(FAILED);
  }
  if (write_hist(fp, feature->hist2, feature->pts, HIST2_BINS) == FAILED) {
    fprintf(stderr, "Error writing histogram B of %dpts\n", feature->pts);
    return(FAILED);
  }
  if (write_hist(fp, feature->hist3, feature->pts, HIST3_BINS) == FAILED) {
    fprintf(stderr, "Error writing histogram G of %dpts\n", feature->pts);
    return(FAILED);
  }
  if (write_hist(fp, feature->hist4, feature->pts, HIST4_BINS) == FAILED) {
    fprintf(stderr, "Error writing histogram G of %dpts\n", feature->pts);
    return(FAILED);
  }
  if (write_hist(fp, feature->hist5, feature->pts, HIST5_BINS) == FAILED) {
    fprintf(stderr, "Error writing histogram G of %dpts\n", feature->pts);
    return(FAILED);
  }
  if (write_hist(fp, feature->hist6, feature->pts, HIST6_BINS) == FAILED) {
    fprintf(stderr, "Error writing histogram G of %dpts\n", feature->pts);
    return(FAILED);
  }
  if (write_hist(fp, feature->hist7, feature->pts, HIST7_BINS) == FAILED) {
    fprintf(stderr, "Error writing histogram G of %dpts\n", feature->pts);
    return(FAILED);
  }
  if (write_hist(fp, feature->hist8, feature->pts, HIST8_BINS) == FAILED) {
    fprintf(stderr, "Error writing histogram G of %dpts\n", feature->pts);
    return(FAILED);
  }
#ifdef WAVELET
  write_int(fp, feature->peakx, WAVELET_BINS);
  write_int(fp, feature->peaky, WAVELET_BINS);
#endif

  *len    = ftell(fp) - *lstart;

  return(SUCCESSED);
}


int read_feature(struct feature *feature, FILE *fp, int lstart, int len)
{
  int  curpos, i;

  /* read out pyramid in binary format */
  curpos = ftell(fp);
  if (curpos != lstart) fseek(fp, lstart, SEEK_SET);

  read_int(fp, &(feature->pts),    1);
  read_int(fp, &(feature->status), 1);
  if (read_hist(fp, feature->hist0, feature->pts, HIST0_BINS) == FAILED) 
    return(FAILED);
  if (read_hist(fp, feature->hist1, feature->pts, HIST1_BINS) == FAILED) 
    return(FAILED);
  if (read_hist(fp, feature->hist2, feature->pts, HIST2_BINS) == FAILED) 
    return(FAILED);
  if (read_hist(fp, feature->hist3, feature->pts, HIST3_BINS) == FAILED) 
    return(FAILED);
  if (read_hist(fp, feature->hist4, feature->pts, HIST4_BINS) == FAILED) 
    return(FAILED);
  if (read_hist(fp, feature->hist5, feature->pts, HIST5_BINS) == FAILED) 
    return(FAILED);
  if (read_hist(fp, feature->hist6, feature->pts, HIST6_BINS) == FAILED) 
    return(FAILED);
  if (read_hist(fp, feature->hist7, feature->pts, HIST7_BINS) == FAILED) 
    return(FAILED);
  if (read_hist(fp, feature->hist8, feature->pts, HIST8_BINS) == FAILED) 
    return(FAILED);

#ifdef WAVELET
  read_int(fp, feature->peakx, WAVELET_BINS);
  read_int(fp, feature->peaky, WAVELET_BINS);
#endif

  return(SUCCESSED);
}


struct feature *new_feature(void)
{
  struct feature    *feature;

  /* Initialize set of location nodes for each image */
  feature = (struct feature *) malloc(sizeof(struct feature));
  reset_feature(feature);

  return(feature);
}


int dump_feature(struct feature *feature)
{
  int    l;

  /* read out pyramid in binary format */
  fprintf(stdout, "%d %d %d %d\n", feature->status, feature->pts, 
                                   feature->row, feature->col);
  for (l=0; l<HIST0_BINS; l++) fprintf(stdout, "%g ", feature->hist0[l]);
  fprintf(stdout, "\n");
  for (l=0; l<HIST1_BINS; l++) fprintf(stdout, "%g ", feature->hist1[l]);
  fprintf(stdout, "\n");
  for (l=0; l<HIST2_BINS; l++) fprintf(stdout, "%g ", feature->hist2[l]);
  fprintf(stdout, "\n");
  for (l=0; l<HIST3_BINS; l++) fprintf(stdout, "%g ", feature->hist3[l]);
  fprintf(stdout, "\n");
  for (l=0; l<HIST4_BINS; l++) fprintf(stdout, "%g ", feature->hist4[l]);
  fprintf(stdout, "\n");
  for (l=0; l<HIST5_BINS; l++) fprintf(stdout, "%g ", feature->hist5[l]);
  fprintf(stdout, "\n");
  for (l=0; l<HIST6_BINS; l++) fprintf(stdout, "%g ", feature->hist6[l]);
  fprintf(stdout, "\n");
  for (l=0; l<HIST7_BINS; l++) fprintf(stdout, "%g ", feature->hist7[l]);
  fprintf(stdout, "\n");
  for (l=0; l<HIST8_BINS; l++) fprintf(stdout, "%g ", feature->hist8[l]);
  fprintf(stdout, "\n");

  return(SUCCESSED);
}

int reset_feature(struct feature *feature)
{
  int    l;

  feature->row = 0;
  feature->col = 0;
  feature->status = 0;
  feature->pts = 0;
  for (l=0; l<HIST0_BINS; l++) feature->hist0[l] = 0;
  for (l=0; l<HIST1_BINS; l++) feature->hist1[l] = 0;
  for (l=0; l<HIST2_BINS; l++) feature->hist2[l] = 0;
  for (l=0; l<HIST3_BINS; l++) feature->hist3[l] = 0;
  for (l=0; l<HIST4_BINS; l++) feature->hist4[l] = 0;
  for (l=0; l<HIST5_BINS; l++) feature->hist5[l] = 0;
  for (l=0; l<HIST6_BINS; l++) feature->hist6[l] = 0;
  for (l=0; l<HIST7_BINS; l++) feature->hist7[l] = 0;
  for (l=0; l<HIST8_BINS; l++) feature->hist8[l] = 0;
#ifdef WAVELET
  for (l=0; l<WAVELET_BINS; l++) feature->peakx[l] = 0;
  for (l=0; l<WAVELET_BINS; l++) feature->peaky[l] = 0;
#endif

  return(SUCCESSED);
}

int copy_feature(struct feature *feature, struct feature *newfeature)
{
  int    l;

  feature->row = newfeature->row;
  feature->col = newfeature->col;
  feature->pts = newfeature->pts;
  feature->status = newfeature->status;
  for (l=0; l<HIST0_BINS; l++) feature->hist0[l]=newfeature->hist0[l];
  for (l=0; l<HIST1_BINS; l++) feature->hist1[l]=newfeature->hist1[l];
  for (l=0; l<HIST2_BINS; l++) feature->hist2[l]=newfeature->hist2[l];
  for (l=0; l<HIST3_BINS; l++) feature->hist3[l]=newfeature->hist3[l];
  for (l=0; l<HIST4_BINS; l++) feature->hist4[l]=newfeature->hist4[l];
  for (l=0; l<HIST5_BINS; l++) feature->hist5[l]=newfeature->hist5[l];
  for (l=0; l<HIST6_BINS; l++) feature->hist6[l]=newfeature->hist6[l];
  for (l=0; l<HIST7_BINS; l++) feature->hist7[l]=newfeature->hist7[l];
  for (l=0; l<HIST8_BINS; l++) feature->hist8[l]=newfeature->hist8[l];
#ifdef WAVELET
  for (l=0; l<WAVELET_BINS; l++) feature->peakx[l] = newfeature->peakx[l];
  for (l=0; l<WAVELET_BINS; l++) feature->peaky[l] = newfeature->peaky[l];
#endif

  return(SUCCESSED);
}

int add_feature(struct feature *feature, struct feature *newfeature, float ratio)
{
  int    l, pts1, pts2, pts;

  /* read out pyramid in binary format */
  feature->status = newfeature->status;
  pts1 = feature->pts;
  pts2 = (int) (newfeature->pts * ratio);
  pts  = pts1 + pts2;
  for (l=0; l<HIST0_BINS; l++) 
    feature->hist0[l]=(feature->hist0[l]*pts1+newfeature->hist0[l]*pts2)/pts;
  for (l=0; l<HIST1_BINS; l++) 
    feature->hist1[l]=(feature->hist1[l]*pts1+newfeature->hist1[l]*pts2)/pts;
  for (l=0; l<HIST2_BINS; l++)
    feature->hist2[l]=(feature->hist2[l]*pts1+newfeature->hist2[l]*pts2)/pts;
  for (l=0; l<HIST3_BINS; l++) 
    feature->hist3[l]=(feature->hist3[l]*pts1+newfeature->hist3[l]*pts2)/pts;
  for (l=0; l<HIST4_BINS; l++) 
    feature->hist4[l]=(feature->hist4[l]*pts1+newfeature->hist4[l]*pts2)/pts;
  for (l=0; l<HIST5_BINS; l++)
    feature->hist5[l]=(feature->hist5[l]*pts1+newfeature->hist5[l]*pts2)/pts;
  for (l=0; l<HIST6_BINS; l++)
    feature->hist6[l]=(feature->hist6[l]*pts1+newfeature->hist6[l]*pts2)/pts;
  for (l=0; l<HIST7_BINS; l++)
    feature->hist7[l]=(feature->hist7[l]*pts1+newfeature->hist7[l]*pts2)/pts;
  for (l=0; l<HIST8_BINS; l++)
    feature->hist8[l]=(feature->hist8[l]*pts1+newfeature->hist8[l]*pts2)/pts;
#ifdef WAVELET
  for (l=0; l<WAVELET_BINS; l++) feature->peakx[l] = newfeature->peakx[l];
  for (l=0; l<WAVELET_BINS; l++) feature->peaky[l] = newfeature->peaky[l];
#endif
  feature->pts += newfeature->pts;

  return(SUCCESSED);
}

int adjust_feature(struct feature *feature, struct feature *newfeature, float ratio)
{
  int    l;

  /* read out pyramid in binary format */
  feature->pts = max(newfeature->pts, 10000);
  feature->status = newfeature->status;
  for (l=0; l<HIST0_BINS; l++) 
    feature->hist0[l]=(feature->hist0[l]*(1-ratio)+newfeature->hist0[l]*ratio);
  for (l=0; l<HIST1_BINS; l++) 
    feature->hist1[l]=(feature->hist1[l]*(1-ratio)+newfeature->hist1[l]*ratio);
  for (l=0; l<HIST2_BINS; l++)
    feature->hist2[l]=(feature->hist2[l]*(1-ratio)+newfeature->hist2[l]*ratio);
  for (l=0; l<HIST3_BINS; l++) 
    feature->hist3[l]=(feature->hist3[l]*(1-ratio)+newfeature->hist3[l]*ratio);
  for (l=0; l<HIST4_BINS; l++) 
    feature->hist4[l]=(feature->hist4[l]*(1-ratio)+newfeature->hist4[l]*ratio);
  for (l=0; l<HIST5_BINS; l++)
    feature->hist5[l]=(feature->hist5[l]*(1-ratio)+newfeature->hist5[l]*ratio);
  for (l=0; l<HIST6_BINS; l++)
    feature->hist6[l]=(feature->hist6[l]*(1-ratio)+newfeature->hist6[l]*ratio);
  for (l=0; l<HIST7_BINS; l++)
    feature->hist7[l]=(feature->hist7[l]*(1-ratio)+newfeature->hist7[l]*ratio);
  for (l=0; l<HIST8_BINS; l++)
    feature->hist8[l]=(feature->hist8[l]*(1-ratio)+newfeature->hist8[l]*ratio);
#ifdef WAVELET
  for (l=0; l<WAVELET_BINS; l++) feature->peakx[l] = newfeature->peakx[l];
  for (l=0; l<WAVELET_BINS; l++) feature->peaky[l] = newfeature->peaky[l];
#endif

  return(SUCCESSED);
}

FILE *open_database(char *dbname, char *subdir, char *mode)
{

  char longname[256];
  FILE *fp;

#if defined(UNIX) || defined(WIN32)
  sprintf(longname, "%s/%s/db_feature",  dbname, subdir);
#else
  sprintf(longname, ":%s:%s:db_feature", dbname, subdir);
#endif

  if ((fp = fopen(longname, mode)) == NULL) {
    fprintf(stderr, "Error: Cannot open database %s\n", longname);
    exit(1);
  }

  return(fp);

}

void close_database(FILE *fp)
{

  fclose(fp);

}

void free_feature(struct feature *feature)
{

  free((void *)feature);

}


static int Gauss_smoothing(struct float_img *data_img)
{
  int   i, j, l;
  int   i1, j1, m, n;
  float	***color1, ***color2;
  float sum;

  color1 = data_img->color; 
  color2 = (float ***) malloc (sizeof (float **) * 3);
  for (l=0; l<3; l++)
    color2[l] = (float **) get_img(data_img->col, data_img->row, sizeof(float));
  
  for (m=0, sum=0; m<5; m++) for (n=0; n<5; n++) sum += filter[m][n];
  for (m=0       ; m<5; m++) for (n=0; n<5; n++) filter[m][n] /= sum;

  for (i=0; i<data_img->row; i++)
  for (j=0; j<data_img->col; j++) {
    for (l=0; l<3; l++) color2[l][i][j] = 0.0;
    for (m=0; m<5; m++) {
      for (n=0; n<5; n++) {
        i1 = i+m-2;
        if (i1 < 0) i1 = -i1;
        if (i1 >= data_img->row) i1 = 2*data_img->row-i1-1;
        j1 = j+n-2;
        if (j1 < 0) j1 = -j1;
        if (j1 >= data_img->col) j1 = 2*data_img->col-j1-1;
        for (l=0; l<3; l++) color2[l][i][j] += color1[l][i1][j1]*filter[m][n];
      }
    }
  }

  data_img->color = color2;
  for (l=0; l<3; l++) free_img((void **) color1[l]);
  free( (void *) color1 );

  return(0);

}

static int color_to_texture(struct float_img *data_img, struct float_img *texture)
{
  int   i, j, l, scale;
  int   i1, i2, i3, j1, j2, j3;
  float	***color1,***color2;
  float gx, gy, value;

  scale  = 1;
  color1 = data_img->color; 

  texture->row = data_img->row;
  texture->col = data_img->col;
  texture->color = color2 = (float ***) malloc (sizeof (float **) * 3);
  for (l=0; l<3; l++)
    texture->color[l] = (float **) 
      get_img(texture->col, texture->row, sizeof(float));
  
  for(i=0; i<data_img->row; i++)
  for(j=0; j<data_img->col; j++) {
    for(l=0; l<3; l++) color2[l][i][j] = 0.0;
    if (i < scale || i >= data_img->row-scale ||
        j < scale || j >= data_img->col-scale) continue;
    i1 = i-scale;
    if (i1 < 0) i1 = -i1;
    i2 = i;
    i3 = i+scale;
    if (i3 >= data_img->row) i3 = 2*data_img->row-i3-1;
    j1 = j-scale;
    if (j1 < 0) j1 = -j1;
    j2 = j;
    j3 = j+scale;
    if (j3 >= data_img->col) j3 = 2*data_img->col-j3-1;
    /* Sobel operator */
    for(l=0; l<3; l++) {
      gx = ((color1[l][i1][j3] - color1[l][i1][j1]) +
            (color1[l][i2][j3] - color1[l][i2][j1]) * 2 +
            (color1[l][i3][j3] - color1[l][i3][j1])) * 0.25;
      gy = ((color1[l][i3][j1] - color1[l][i1][j1]) +
            (color1[l][i3][j2] - color1[l][i1][j2]) * 2 +
            (color1[l][i3][j3] - color1[l][i1][j3])) * 0.25;
      value = sqrt(gx * gx + gy * gy);
      color2[l][i][j] = value;
    }
  }

  return(0);

}

static int color_to_edgemap(struct float_img *data_img, struct float_img *edgemap)
{
  struct float_img smooth_img;
  int   i, j, l, index, scale;
  int   i1, i2, i3, j1, j2, j3;
  int   ratio[3][256];
  float	***color1,***color2, ***color3;
  float gx, gy, value, angle;
  float maxgrad[3], edge[3], sum, sum1[3], sum2[3];

  smooth_img.row = data_img->row;
  smooth_img.col = data_img->col;
  smooth_img.color = (float ***) malloc (sizeof (float **) * 3);
  color1 = smooth_img.color;
  for (l=0; l<3; l++)
    smooth_img.color[l] = (float **) 
      get_img(data_img->col, data_img->row, sizeof(float));

  for (i=0; i<data_img->row; i++)
  for (j=0; j<data_img->col; j++) 
    for (l=0; l<3; l++) color1[l][i][j] = data_img->color[l][i][j];

  for (l=0; l<2; l++) Gauss_smoothing(&smooth_img);

  color1 = smooth_img.color;
  color2 = (float ***) malloc (sizeof (float **) * 3);
  for (l=0; l<3; l++)
    color2[l] = (float **) 
      get_img(data_img->col, data_img->row, sizeof(float));
  
  edgemap->row = data_img->row;
  edgemap->col = data_img->col;
  edgemap->color = color3 = (float ***) malloc (sizeof (float **) * 3);
  for (l=0; l<3; l++)
    edgemap->color[l] = (float **) 
      get_img(edgemap->col, edgemap->row, sizeof(float));
  
  scale = 4;
  for (l=0; l<3; l++) maxgrad[l] = sum1[l] = sum2[l] = 0;
  for (l=0; l<3; l++) for (i=0; i<256; i++) ratio[l][i] = 0;
  for(i=0; i<data_img->row; i++)
  for(j=0; j<data_img->col; j++) {
    for(l=0; l<3; l++) {
      color2[l][i][j] = 0.0;
      color3[l][i][j] = 0.0;
    }
    if (i < scale || i >= data_img->row-scale ||
        j < scale || j >= data_img->col-scale) continue;
    i1 = i-scale;
    if (i1 < 0) i1 = -i1;
    i2 = i;
    i3 = i+scale;
    if (i3 >= data_img->row) i3 = 2*data_img->row-i3-1;
    j1 = j-scale;
    if (j1 < 0) j1 = -j1;
    j2 = j;
    j3 = j+scale;
    if (j3 >= data_img->col) j3 = 2*data_img->col-j3-1;
    /* Sobel operator */
    for(l=0; l<3; l++) {
      gx = ((color1[l][i1][j3] - color1[l][i1][j1]) +
            (color1[l][i2][j3] - color1[l][i2][j1]) * 2 +
            (color1[l][i3][j3] - color1[l][i3][j1])) * 0.25;
      gy = ((color1[l][i3][j1] - color1[l][i1][j1]) +
            (color1[l][i3][j2] - color1[l][i1][j2]) * 2 +
            (color1[l][i3][j3] - color1[l][i1][j3])) * 0.25;
      value = sqrt(gx * gx + gy * gy);
      if (gx == 0 && gy == 0) angle = 0.0;
      else                    angle = atan2(gy, gx);
      color2[l][i][j] = value;
      color3[l][i][j] = angle;
      sum1[l] += value;
      sum2[l] += value * value;

      if (maxgrad[l] < value) maxgrad[l] = value;
      index = (int) max(0, min(255, value)); 
      ratio[l][index]++;
    }
  }

  for (l=0; l<3; l++) edge[l] = 0;
  for (l=0; l<3; l++) {
    sum1[l] /= (data_img->col * data_img->row);
    sum2[l] /= (data_img->col * data_img->row);
    edge[l] = sqrt(sum2[l]);
    for (i=255, sum=0; i>=0 && i > edge[l]; i--) sum += ratio[l][i];
    sum     /= (data_img->col * data_img->row);
  }

  for(i=0; i<data_img->row; i++)
  for(j=0; j<data_img->col; j++) {
    /* Sobel operator */
    for(l=0; l<3; l++) {
      if (color2[l][i][j] < edge[l]) 
        color3[l][i][j] = - 2 * PI;
      else
        color3[l][i][j] = color3[l][i][j];
    }
  }
  
  for (l=0; l<3; l++) free_img((void **) color1[l]);
  free( (void *) color1 );
  for (l=0; l<3; l++) free_img((void **) color2[l]);
  free( (void *) color2 );

  return(0);

}

static int color_to_gradient(struct float_img *data_img, 
	 	                     struct float_img *grad1_img,
		                     struct float_img *grad2_img)
{
  int   i, j, l, index;
  int   i1, i2, i3, j1, j2, j3;
  int   ratio[3][256];
  float	***color1,***color2, ***color3;
  float gx, gy, value, angle;
  float maxgrad[3], edge[3], sum, sum1[3], sum2[3];

  grad1_img->row = data_img->row;
  grad1_img->col = data_img->col;
  grad1_img->color = (float ***) malloc (sizeof (float **) * 3);
  for (l=0; l<3; l++)
    grad1_img->color[l] = (float **) 
      get_img(grad1_img->col, grad1_img->row, sizeof(float));
  
  grad2_img->row = data_img->row;
  grad2_img->col = data_img->col;
  grad2_img->color = (float ***) malloc (sizeof (float **) * 3);
  for (l=0; l<3; l++)
    grad2_img->color[l] = (float **) 
      get_img(grad2_img->col, grad2_img->row, sizeof(float));
  
  color1 = data_img->color; 
  color2 = grad1_img->color; 
  color3 = grad2_img->color;
  for (l=0; l<3; l++) maxgrad[l] = sum1[l] = sum2[l] = 0;
  for (l=0; l<3; l++) for (i=0; i<256; i++) ratio[l][i] = 0;
  for(i=0; i<data_img->row; i++)
  for(j=0; j<data_img->col; j++) {
    for(l=0; l<3; l++) {
      color2[l][i][j] = 0.0;
      color3[l][i][j] = 0.0;
    }
    if (i==0 || j==0 || i==data_img->row-1 || j==data_img->col-1) continue;
    i1 = max(0, i-1); i2 = i; i3 = min(data_img->row-1, i+1);
    j1 = max(0, j-1); j2 = j; j3 = min(data_img->col-1, j+1);
    /* Sobel operator */
    for(l=0; l<3; l++) {
      gx = ((color1[l][i1][j3] - color1[l][i1][j1]) +
            (color1[l][i2][j3] - color1[l][i2][j1]) * 2 +
            (color1[l][i3][j3] - color1[l][i3][j1])) * 0.25;
      gy = ((color1[l][i3][j1] - color1[l][i1][j1]) +
            (color1[l][i3][j2] - color1[l][i1][j2]) * 2 +
            (color1[l][i3][j3] - color1[l][i1][j3])) * 0.25;
      value = sqrt(gx * gx + gy * gy);
      if (gx == 0 && gy == 0) angle = 0.0;
      else                    angle = atan2(gy, gx);
      color2[l][i][j] = value;
      color3[l][i][j] = angle;
      sum1[l] += value;
      sum2[l] += value * value;

      if (maxgrad[l] < value) maxgrad[l] = value;
      index = (int) max(0, min(255, value)); 
      ratio[l][index]++;
    }
  }

  for (l=0; l<3; l++) edge[l] = 0;
  for (l=0; l<3; l++) {
    sum1[l] /= (data_img->col * data_img->row);
    sum2[l] /= (data_img->col * data_img->row);
    edge[l] = sqrt(sum2[l]);
    for (i=255, sum=0; i>=0 && i > edge[l]; i--) sum += ratio[l][i];
    sum     /= (data_img->col * data_img->row);
  }

  for(i=0; i<data_img->row; i++)
  for(j=0; j<data_img->col; j++) {
    /* Sobel operator */
    for(l=0; l<3; l++) {
      if (color2[l][i][j] < edge[l]) 
        color3[l][i][j] = - 2 * PI;
      else
        color3[l][i][j] = color3[l][i][j];
    }
  }
  
  return(0);

}

int color_to_laplacian(struct float_img *data_img, struct float_img *lapl_img)
{
  int   i, j, l;
  int   i1, i2, j1, j2;
  float	***color1,***color2;

  lapl_img->row = data_img->row;
  lapl_img->col = data_img->col;
  lapl_img->color = (float ***) malloc (sizeof (float **) * 3);
  for (l=0; l<3; l++)
    lapl_img->color[l] = (float **) 
      get_img(lapl_img->col, lapl_img->row, sizeof(float));
  
  color1 = data_img->color; color2 = lapl_img->color;
  for(i2=0; i2<data_img->row; i2++)
  for(j2=0; j2<data_img->col; j2++) {
    for(l=0; l<3; l++) color2[l][i2][j2] = 0; 
    for (i=0; i<MASKSIZE; i++) 
    for (j=0; j<MASKSIZE; j++) {
      i1 = i2 + i - (MASKSIZE-1)/2;
      j1 = j2 + j - (MASKSIZE-1)/2;
      if (i1 < 0) i1 = -i1;
      if (j1 < 0) j1 = -j1;
      if (i1 > data_img->row-1) i1 = 2 * data_img->row - i1 - 1;
      if (j1 > data_img->col-1) j1 = 2 * data_img->col - j1 - 1;
      for(l=0; l<3; l++) 
        color2[l][i2][j2] += color1[l][i1][j1] * mask[i][j];
    }
  }

  return(0);

}


static int color_to_wavelet(struct float_img *data_img, struct float_img *wavelet)
{
  int   i, j, l, xoffset, yoffset;


/*
  wavelet->row = min(data_img->row, data_img->col);
  wavelet->col = min(data_img->row, data_img->col);
*/
  wavelet->row = data_img->row;
  wavelet->col = data_img->col;
  wavelet->color = (float ***) malloc (sizeof (float **) * 3);
  for (l=0; l<3; l++)
    wavelet->color[l] = (float **) 
      get_img(wavelet->col, wavelet->row, sizeof(float));

  xoffset = (data_img->col - wavelet->col) / 2;
  yoffset = (data_img->row - wavelet->row) / 2;
  for (l=0; l<3; l++) 
  for (i=0; i<wavelet->row; i++)
  for (j=0; j<wavelet->col; j++)
	  wavelet->color[l][i][j] = data_img->color[l][i+yoffset][j+xoffset];
  
  for (l=0; l<3; l++)
	wavelet2(wavelet->color[l], wavelet->color[l], wavelet->col, wavelet->row, haar, 2);

  return(0);

}


/*                                                                      */
/* This Program accept a 2-D signal input whose size are  power of two. */
/* And Recursive do a octive structive QMF, aks wavelet decomposition.  */
/*                                                                      */


static void wavelet2(
					float	**data,
					float   **output,
					int		width,
					int		height,
				    float 	*filt,  	      /*  QMF filter coefficients  */
					int 	flen) 		      /*  QMF filter length  */
{
  int	p, power, w, h;

  power = (int) (log((double) min(width, height)) / log(2.0));
  w = width; h = height;
  anqmf2(data, output, w, h, filt, flen);
  w /= 2;
  h /= 2;
  for (p=1; p<power; p++) {
	anqmf2(output, output, w, h, filt, flen);
	w /= 2;
	h /= 2;
  }

}



/*                                                                      */
/* This Program accept a 1-D signal input whose length is power of two. */
/* And Recursive do a octive structive QMF, aks wavelet decomposition.  */
/*                                                                      */

static void wavelet(float	*data,
					float   *output,
					int		dlen,
					float 	*filt,  	      /*  QMF filter coefficients  */
					int 	flen) 		      /*  QMF filter length  */
{
  int	p, power, size;

  power = (int) (log((double) dlen) / log(2.0));
  size = dlen;
  anqmf(data, output, size, filt, flen);
  size /= 2;
  for (p=1; p<power; p++) {
	anqmf(output, output, size, filt, flen);
	size /= 2;
  }
 
}


/*                                                                      */
/* This Program accept a 2-D signal input, do a one time decomposition.	*/
/*                                                                      */

static void anqmf2(
			    float	**data,
				float	**output,
				int		width,
				int		height,
				float 	*filt,  	      /*  QMF filter coefficients  */
				int 	flen) 		      /*  QMF filter length  	   */
{
  int	i, j;
  float buf[256];

  for (i=0; i<height; i++) {
    anqmf(data[i], output[i], width, filt, flen);
  }
  for (i=0; i<width; i++) {
    for (j=0; j<height; j++) buf[j] = output[j][i];
    anqmf(buf, buf, height, filt, flen);
    for (j=0; j<height; j++) output[j][i] = buf[j];
  }
}



/*                                                                      */
/* This Program accept a 1-D signal input, do a one time decomposition. */
/*                                                                      */

static void anqmf(
            float	*data,
            float	*output,
            int		dlen,
            float 	*filt,  	      /*  QMF filter coefficients  */
            int 	flen) 		      /*  QMF filter length  */
{
  int	i, j, k;
  float lowband[256], highband[256], f1[256], f2[256], sign, sum;

  /* First create two orthonormal filters */
  for (i=0, sum=0; i<flen; i++) sum += filt[i] * filt[i];
  for (i=0; i<flen; i++) f1[i] = filt[i] / sqrt((double) sum);
  for (i=flen/2-1, sign=1;  i>=0;   i--) { f2[i] = f1[i] * sign; sign *= -1;}
  for (i=flen/2,   sign=-1; i<flen; i++) { f2[i] = f1[i] * sign; sign *= -1;}

  /* pass signal through filter bank */
  for (i=0; i<dlen/2; i++)
  {
    highband[i] = 0.0;
    lowband[i]  = 0.0;
    for (j=0; j<flen; j++)
    {
      k = i * 2 + (flen/2 - j);
      if      (k <  0) k = -k;
      else if (k >= dlen) k = 2 * dlen - 1 - k;
      lowband[i]   += f1[j] * data[k];
      highband[i]  += f2[j] * data[k];
    }
  }
  for (i=0;      i<dlen/2; i++) output[i]  = lowband[i];
  for (i=dlen/2; i<dlen;   i++) output[i]  = highband[i-dlen/2];

}


