/* The TIFF writer contained in this code */
/* assumed that the machine on which it   */
/* is running stores numbers with the     */
/* most-significant byte first, i.e., in  */
/* the "big-endian" byte-order.           */

#include  "tiff.h"

#define LongReverse(x) \
        ((unsigned long int)((((unsigned long int)(x) & 0x000000ffU) << 24) | \
                             (((unsigned long int)(x) & 0x0000ff00U) <<  8) | \
                             (((unsigned long int)(x) & 0x00ff0000U) >>  8) | \
                             (((unsigned long int)(x) & 0xff000000U) >> 24)))
#define ShortReverse(x) \
        ((unsigned short int)((((unsigned short int)(x) & 0x00ffU) << 8) | \
                              (((unsigned short int)(x) & 0xff00U) >> 8)))


static struct Rational {
  unsigned long Numer;
  unsigned long Denom;
};

static union TIFF_value {		/* don't know the type of the     */
  unsigned char UChar;			/* data value(s) in advance       */
  unsigned short UShort;
  unsigned long ULong;
  struct Rational Rat;

  unsigned char *UCharArray;
  unsigned short *UShortArray;
  unsigned long *ULongArray;
  struct Rational *RatArray;
};

static struct TIFF_field {
  unsigned short Tag;			/* tag for TIFF field             */
  unsigned short Type;			/* type for values in field       */
  unsigned long  Count;			/* number of values in field      */
  union TIFF_value Value;		/* the ValueOrOffset word in the  */
					/* field contains the Value       */
					/* instead of pointing to the     */
					/* Value if and only if the Value */
					/* fits into 4 bytes; in this     */
					/* case the value is stored in    */
					/* the lower-numbered bytes;      */
					/* if value does not fit into 4   */
					/* bytes, the entry there is the  */
					/* offset (wrt beginning of file) */
					/* to the value                   */
  int SizeOfType;			/* number of bytes per element of */
					/* the Value                      */
};

static struct IFD {
  unsigned short NumberOfFields;	/* number of IFD entries for this */
					/* image                          */
  struct TIFF_field *Fields;		/* this array should have         */
					/* NumberOfFields elements        */
};

static struct TIFF_header {
  unsigned short ByteOrder;		/* for interpreting multi-byte    */
					/* numbers                        */
					/* 0x4949 = little-endian         */
					/* 0x4D4D = big-endian            */

  unsigned short FortyTwo;		/* the second word in a TIFF      */
					/* image header is always 42      */

  unsigned long OffsetOfFirstIFD; 	/* offset of first Image File     */
                                        /* Directory (IFD) (wrt beginning */
					/* of file; i.e., in bytes,       */
					/* counting the first byte of     */
					/* the file as number 0)          */
};

static struct DataLocation {
  unsigned long num_strips;
  unsigned long rows_per_strip;
  unsigned long *strip_offsets;
  unsigned long *strip_byte_counts;
  unsigned long offset_of_byte_after_data;
  unsigned long bytes_per_row;
};

/* subroutines */
static int tiffHostType(int dummy);
static int FreeIFD ( struct IFD *ifd );
static int FreeFieldValues ( struct TIFF_field *field );
static int WriteHeaderAndIFD ( FILE *fp, struct IFD *ifd,
                               struct TIFF_header *header );
static int WriteIFD ( FILE *fp, struct IFD *ifd,
                      struct TIFF_header *header );
static int WriteField ( FILE *fp, struct TIFF_field *field,
                        unsigned long starting_offset,
                        unsigned long *outside_IFD_offset );
static int WriteValue ( FILE *fp, struct TIFF_field *field,
                        unsigned long *outside_IFD_offset );
static int WriteArrayOfValues ( FILE *fp, struct TIFF_field *field );
static int WriteSingleValue ( FILE *fp, struct TIFF_field *field );
static int WriteHeader ( FILE *fp, struct TIFF_header *header );
static int WriteRational ( FILE *fp, struct Rational *Rat );
static int WriteUnsignedLong ( FILE *fp, unsigned long *UnsignedLong );
static int WriteUnsignedShort ( FILE *fp, unsigned short *UnsignedShort );
static int WriteUnsignedChar ( FILE *fp, unsigned char *UnsignedChar );
static int WriteImageData ( FILE *fp, struct TIFF_img *img,
                            struct DataLocation *DataLoc );
static int WriteUncompressedImageData ( FILE *fp, struct TIFF_img *img,
                                        struct DataLocation *DataLoc );
static int WriteUncompressedStrip ( FILE *fp, struct TIFF_img *img,
                                    struct DataLocation *DataLoc,
                                    unsigned long strip_index );
static int WriteUncompGrayscaleRow ( FILE *fp, struct TIFF_img *img,
                                     struct DataLocation *DataLoc,
                                     unsigned long row_index );
static int WriteUncompColorRow ( FILE *fp, struct TIFF_img *img,
                                 struct DataLocation *DataLoc,
                                 unsigned long row_index );
static int PrepDataLocStructUncompressed ( struct TIFF_img *img,
                                           struct DataLocation *DataLoc );
static int AllocateDataLocArrays ( struct DataLocation *DataLoc );
static int ComputeNumStrips ( int height, struct DataLocation *DataLoc );
static int ComputeRowsPerStrip ( struct TIFF_img *img,
                                 struct DataLocation *DataLoc );
static int PrepareHeaderAndIFD ( struct TIFF_img *img, struct IFD *ifd,
                                 struct TIFF_header *header,
                                 struct DataLocation *DataLoc );
static int PrepareHeader ( struct TIFF_header *header,
                           struct DataLocation *DataLoc );
static int PrepareIFD ( struct TIFF_img *img,
                        struct IFD *ifd,
                        struct DataLocation *DataLoc );
static int AddSpecialFieldEntries ( struct IFD *ifd, char TIFF_type );
static int AddGrayscaleFields ( struct IFD *ifd, char TIFF_type );
static int AddColorFields ( struct IFD *ifd, char TIFF_type );
static int SortFields ( struct IFD *ifd );
static void SwitchFields ( struct TIFF_field *FieldOne,
                           struct TIFF_field *FieldTwo );
static int AddCoreFieldEntries ( struct TIFF_img *img,
                                 struct IFD *ifd,
                                 struct DataLocation *DataLoc );
static int MakeSamplesPerPixelField ( struct IFD *ifd );
static int MakeBitsPerSampleField ( struct IFD *ifd, char TIFF_type );
static int MakeResolutionUnitField ( struct IFD *ifd );
static int MakeResolutionField ( struct IFD *ifd, unsigned short Tag );
static int MakeStripByteCountsField ( struct IFD *ifd,
                                      struct DataLocation *DataLoc );
static int MakeRowsPerStripField ( struct IFD *ifd,
                                   struct DataLocation *DataLoc );
static int MakeStripOffsetsField ( struct IFD *ifd,
                                   struct DataLocation *DataLoc );
static int MakePhotometricInterpretationField ( struct IFD *ifd, char TIFF_type );
static int MakeCompressionField ( struct IFD *ifd, char compress_type );
static int MakeImageWidthOrLengthField ( struct IFD *ifd,
                                         unsigned short Tag,
                                         int dimension );
static unsigned long GetMaxValUL ( unsigned long *array,
                                   unsigned long n_elements );
static int GetImageData ( FILE *fp, struct TIFF_img *img, struct IFD *ifd );
static int CopyImageData ( FILE *fp, struct TIFF_img *img, struct IFD *ifd );
static void FreeDataLocation ( struct DataLocation *DataLoc );
static int CopyUncompressedImageData ( FILE *fp, struct TIFF_img *img,
                                       struct IFD *ifd,
                                       struct DataLocation *DataLoc );
static int GetUncompressedStrip ( FILE *fp, struct TIFF_img *img,
                                  struct IFD *ifd,
                                  struct DataLocation *DataLoc,
                                  unsigned long strips_already_read );
static int GetUncompressedColorRow ( FILE *fp, struct TIFF_img *img,
                                     unsigned long current_row );
static int GetUncompressedGrayscaleRow ( FILE *fp, struct TIFF_img *img,
                                         unsigned long current_row );
static int IsThereAnyMoreDataInThisStrip ( struct IFD *ifd, int width,
                                           struct DataLocation *DataLoc,
                                           unsigned long current_strip_index,
                                           unsigned long rows_already_read );
static int ComputeBytesPerRow ( struct IFD *ifd, int width,
                                unsigned long *BytesPerRow );
static int ComputeBytesPerSample ( struct IFD *ifd,
                                   unsigned long *BytesPerSample );
static int ComputeTotalBitsPerSample ( struct IFD *ifd, unsigned long *Total );
static int GetSamplesPerPixel ( struct IFD *ifd, unsigned short *Number );
static int GetCompressionValue ( struct IFD *ifd, unsigned short *Value );
static int GetImageDataLocationInfo ( struct IFD *ifd,
                                      struct DataLocation *DataLoc );
static int GetStripOffsets ( struct IFD *ifd,
                             struct DataLocation *DataLoc );
static int GetStripByteCounts ( struct IFD *ifd,
                                struct DataLocation *DataLoc );
static int PutValuesIntoULongArray ( struct TIFF_field *field,
                                     unsigned long **array,
                                     unsigned long num_elements );
static int GetNumberOfStrips ( struct IFD *ifd,
                               unsigned long *num_strips );
static int GetRowsPerStrip ( struct IFD *ifd,
                             unsigned long *rows_per_strip );
static void WrongValueType ( char *FunctionName, unsigned short Tag,
                             unsigned short Type );
static int AllocateImageDataArray ( struct TIFF_img *img );
static int GetHeightAndWidth ( struct IFD *ifd, int *height, int *width );
static int GetCompression ( struct IFD *ifd, char *compress_type );
static int VerifyIFD ( struct IFD *ifd, char *TIFF_type );
static int GetImageType ( struct IFD *ifd, char *TIFF_type );
static int IsImageFullColor ( struct IFD *ifd );
static int IsImageGrayscale ( struct IFD *ifd );
static struct TIFF_field *GetFieldStructure ( struct IFD *ifd,
                                              unsigned short Tag );
static int CheckForCoreFields ( struct IFD *ifd );
static int WhatAboutCoreField ( struct IFD *ifd, unsigned short Tag );
static int AddDefaultField ( struct IFD *ifd, unsigned short Tag );
static void TellUserACoreFieldIsNecessary ( unsigned short Tag );
static int IsThereADefaultFor ( unsigned short Tag );
static int IsThereAFieldFor ( struct IFD *ifd, unsigned short Tag );
static int ReadIFD ( FILE *fp, struct IFD *ifd,
                     struct TIFF_header *header );
static int SeeIfThereAreOtherIFDs ( FILE *fp, unsigned long OffsetOfIFD,
                                    unsigned short MaxNumberOfFields );
static int SetFilePositionAtFirstByteOfIthField ( FILE *fp,
                                                  unsigned long OffsetOfIFD,
                                                  unsigned short i );
static int CopyFieldIfTagIsRecognized ( FILE *fp, struct IFD *ifd );
static int AllocateAndCopyField ( FILE *fp, struct IFD *ifd,
                                  unsigned short Tag, unsigned short Type );
static int IsTypeExpectedWithTag ( unsigned short Tag, unsigned short Type );
static int IsTypeRecognized ( unsigned short Type );
static int CopyField ( FILE *fp, struct TIFF_field *field,
                       unsigned short Tag, unsigned short Type );
static int GetSizeOfType ( struct TIFF_field *field );
static int CopyValue ( FILE *fp, struct TIFF_field *field );
static int GetArrayOfValues ( FILE *fp, struct TIFF_field *field );
static int GetSingleValue ( FILE *fp, struct TIFF_field *field );
static int AllocateArrayOfValues ( struct TIFF_field *field );
static int CopyCount ( FILE *fp, struct TIFF_field *field );
static int AllocateNewField ( struct IFD *ifd );
static int IsTagRecognized ( unsigned short TempTag );
static int GetRational ( FILE *fp, struct Rational *Rat );
static int GetUnsignedLong ( FILE *fp, unsigned long *UnsignedLong );
static int GetUnsignedShort ( FILE *fp, unsigned short *UnsignedShort );
static int GetUnsignedChar ( FILE *fp, unsigned char *UnsignedChar );
static int GetNumberOfFieldsInInput ( FILE *fp, unsigned long OffsetOfIFD,
                                      unsigned short *NumFieldsInInputFile );
static int SetFilePosition ( FILE *fp, unsigned long position );
static int ReadHeader ( FILE *fp, struct TIFF_header *header );
static int CheckTypeSizes ( void );
static int GetByteOrder ( FILE *fp, struct TIFF_header *header );
static int CheckFortyTwo ( FILE *fp, struct TIFF_header *header );
static int GetOffsetOfFirstIFD ( FILE *fp, struct TIFF_header *header );
static void PrintIFD ( struct IFD *ifd );
static void PrintField ( struct TIFF_field *field );
static void *mget_spc(int num,size_t size);
static void **get_img(int wd,int ht,size_t size);
static void free_img(void **pt);

/* TIFF field tags */
#define	ImageWidth 			256	/* width                  */

#define	ImageLength			257	/* length (height)        */

#define	BitsPerSample			258	/* for image data, number */
						/* of bits per component  */
						/* (4 and 8 are the two   */
						/* allowable values for   */
						/* grayscale and          */ 
						/* palette-color)         */
						/* (for full-color RGB,   */
						/* 8 bits per component;  */
						/* so 8,8,8)              */

#define Compression			259	/* data can be stored     */
#define	NoCompression			1	/* uncompressed or        */
#define	PackBits			32773	/* compressed             */
#define	OneDimModifiedHuffman		2	/* (1DModHuff compression */
						/* is not available for   */
						/* grayscale, palette-    */
						/* color, or full RGB     */
						/* color)                 */

#define	PhotometricInterpretation	262	/* differentiates between */
#define	WhiteIsZero			0	/* white-on-black or      */
#define	BlackIsZero			1	/* black-on-white for     */
						/* binary images,         */
#define	RGB				2	/* full-color RGB, and    */
#define	PaletteColor			3	/* palette-color          */

#define	StripOffsets			273	/* for each strip, the    */
						/* byte offset for that   */
						/* strip (wrt beginning   */
						/* of file)               */

#define	SamplesPerPixel			277	/* number of components   */
						/* per pixel; for full-   */
						/* color RGB this must be */
						/* at least 3             */

#define	RowsPerStrip			278	/* number of rows in each */
 						/* strip (except possibly */
						/* the last strip)        */

#define	StripByteCounts			279	/* for each strip, the    */
						/* number of bytes in     */
						/* that strip, after any  */
						/* compression            */

#define	XResolution			282	/* number of pixels per   */
						/* ResolutionUnit (see    */
						/* above) in ImageWidth   */
						/* direction              */

#define	YResolution			283	/* number of pixels per   */
						/* ResolutionUnit (see    */
						/* above) in ImageLength  */
						/* direction              */

#define	ResolutionUnit			296	/* absolute unit of meas. */
#define	NoAbsoluteUnit			1
#define	Inch				2 	/* (this is the default)  */
#define	Centimeter 			3

#define	ColorMap			320	/* color map for palette- */
						/* color images           */
						/* (the count for this    */
						/* field has to be        */
						/* 3*(2^BitsPerSample))   */

/* TIFF field types */
#define BYTE 		1			/* 1-byte unsigned int    */
#define SHORT 		3			/* 2-byte unsigned int    */
#define LONG 		4			/* 4-byte unsigned int    */
#define RATIONAL	5			/* two LONGS:  first is   */
						/* numerator, second is   */
						/* denominator            */

/* TIFF byte-order possibilities */
#define LittleEndian	0x4949
#define BigEndian	0x4d4d
#define UnknownEndian	0xffff

/* other useful defs */
#define	ERROR		1
#define	NO_ERROR	0

#define	NO		0
#define	YES		1

int TIFF_ORDER, HOST_ORDER;

int write_TIFF ( FILE *fp, struct TIFF_img *img )
{
  struct IFD ifd;
  struct TIFF_header header;
  struct DataLocation DataLoc;

  TIFF_ORDER = BigEndian;
  if ((HOST_ORDER = tiffHostType(0)) == UnknownEndian) return ( ERROR );

  /* ensure that each type will be stored */
  /* in the expected number of bytes      */
  if ( CheckTypeSizes ( ) == ERROR ) return ( ERROR );

  /* write image data, store information about */
  /* location of the data in DataLoc structure */
  if ( WriteImageData ( fp, img, &(DataLoc) ) == ERROR ) 
    return ( ERROR );

  /* prepare header and image file directory */
  if ( PrepareHeaderAndIFD ( img, &(ifd), &(header), &(DataLoc) ) 
       == ERROR ) return ( ERROR );

  /* write header and image file directory */
  if ( WriteHeaderAndIFD ( fp, &(ifd), &(header) ) == ERROR ) 
    return ( ERROR );

  /* free arrays in DataLocation and IFD structures */
  FreeDataLocation ( &(DataLoc) );
  if ( FreeIFD ( &(ifd) ) == ERROR ) return ( ERROR );

  return ( NO_ERROR );
}

void free_TIFF ( struct TIFF_img *img )
{
  unsigned long i;

  if ( img->TIFF_type == 'c' ) {
    for ( i = 0; i < 3; i++ )
      free_img ( ( void ** ) ( img->color[i] ) );
  }
  else free_img ( ( void ** ) ( img->mono ) );
}

static int FreeIFD ( struct IFD *ifd )
{
  unsigned long i;

  /* free array(s) of values */
  for ( i = 0; i < ifd->NumberOfFields; i++ )
    if ( ifd->Fields[i].Count > 1 )
      if ( FreeFieldValues ( &(ifd->Fields[i]) ) == ERROR )
        return ( ERROR );

  /* free IFD fields */
  free ( ( void * ) ifd->Fields );

  return ( NO_ERROR );
}

static int FreeFieldValues ( struct TIFF_field *field )
{
  if ( field->Type == BYTE ) {
    free ( ( void * ) field->Value.UCharArray );
    return ( NO_ERROR );
  }

  if ( field->Type == SHORT ) {
    free ( ( void * ) field->Value.UShortArray );
    return ( NO_ERROR );
  }

  if ( field->Type == LONG ) {
    free ( ( void * ) field->Value.ULongArray );
    return ( NO_ERROR );
  }

  if ( field->Type == RATIONAL ) {
    free ( ( void * ) field->Value.RatArray );
    return ( NO_ERROR );
  }

  fprintf ( stderr, "tiff.c:  function FreeIFD:\n" );
  fprintf ( stderr, "not prepared to free value array of type %d", 
            field->Type );
  return ( ERROR );
}

static int WriteHeaderAndIFD ( FILE *fp, struct IFD *ifd, 
                               struct TIFF_header *header )
{
  /* write header */
  if ( WriteHeader ( fp, header ) == ERROR ) return ( ERROR );

  /* write IFD */
  if ( WriteIFD ( fp, ifd, header ) == ERROR ) return ( ERROR );

  return ( NO_ERROR );
}

static int WriteIFD ( FILE *fp, struct IFD *ifd, 
                      struct TIFF_header *header )
{
  unsigned long outside_IFD_offset, i; 
  unsigned long starting_offset, zero;

  /* compute location of first byte after IFD and four  */
  /* bytes of zeros; this number will be maintained as  */
  /* the address at which field values (which don't fit */
  /* into four bytes) should be written                 */
  outside_IFD_offset = header->OffsetOfFirstIFD + 2 +
                       ( unsigned long ) 
                       ( ifd->NumberOfFields * 12 ) + 4;

  /* set file position */
  if ( SetFilePosition ( fp, header->OffsetOfFirstIFD ) 
       == ERROR ) {
    fprintf ( stderr, "tiff.c:  function WriteIFD:\n" );
    fprintf ( stderr, "error setting file position\n" );
    return ( ERROR );
  }

  /* write number of fields */
  if ( WriteUnsignedShort ( fp, &(ifd->NumberOfFields) ) 
       == ERROR ) return ( ERROR );

  /* loop through fields */
  for ( i = 0; i < ifd->NumberOfFields; i++ ) {

    /* compute starting offset for field */
    starting_offset = header->OffsetOfFirstIFD + 2 + (12 * i);

    /* write field */
    if ( WriteField ( fp, &(ifd->Fields[i]), starting_offset,
                      &(outside_IFD_offset) ) == ERROR ) 
      return ( ERROR );

  }

  /* write four bytes of zero to signify that this */
  /* is that last (only) IFD to have been written  */

  /* set file position */
  starting_offset = header->OffsetOfFirstIFD + 2 + 12 * ifd->NumberOfFields;
  if ( SetFilePosition ( fp, starting_offset ) == ERROR )
    return ( ERROR );
  zero = 0;
  if ( WriteUnsignedLong ( fp, &(zero) ) == ERROR ) 
    return ( ERROR );

  return ( NO_ERROR );
}

static int WriteField ( FILE *fp, struct TIFF_field *field, 
                        unsigned long starting_offset,
                        unsigned long *outside_IFD_offset )
{
  /* set file position at starting_offset */
  if ( SetFilePosition ( fp, starting_offset ) == ERROR )
    return ( ERROR );

  /* write tag */
  if ( WriteUnsignedShort ( fp, &(field->Tag) ) == ERROR )
    return ( ERROR );

  /* write type */
  if ( WriteUnsignedShort ( fp, &(field->Type) ) == ERROR )
    return ( ERROR );

  /* write Count */
  if ( WriteUnsignedLong ( fp, &(field->Count) ) == ERROR )
    return ( ERROR );

  /* write value if value fits into 4   */
  /* bytes; else write offset to value, */
  /* and go to the offset to write down */
  /* the actual field value(s)          */
  if ( WriteValue ( fp, field, outside_IFD_offset ) == ERROR ) 
    return ( ERROR );

  return ( NO_ERROR );
}

static int WriteValue ( FILE *fp, struct TIFF_field *field,
                        unsigned long *outside_IFD_offset )
{
  /* get size of type */
  if ( GetSizeOfType ( field ) == ERROR ) return ( ERROR );

  /* if value will not fit inside four bytes */
  if ( 4 < field->Count * field->SizeOfType ) {

    /* write offset to place where value(s) will be written */
    if ( WriteUnsignedLong ( fp, outside_IFD_offset ) 
         == ERROR ) return ( ERROR );

    /* set file position where value(s) should be written */
    if ( SetFilePosition ( fp, *outside_IFD_offset ) == ERROR ) {
      fprintf ( stderr, "tiff.c:  function WriteValue:\n" );
      fprintf ( stderr, "error setting file position\n" );
      return ( ERROR );
    }

    /* update *outside_IFD_offset; fix it so that the next */
    /* value written outside the ValueOrOffset (the 4-byte */
    /* field entry in IFD) will begin on a word boundary   */
    *outside_IFD_offset += field->Count * field->SizeOfType;
    if ( ( *outside_IFD_offset % 2 ) != 0 ) 
      *outside_IFD_offset = *outside_IFD_offset + 1;
  
  }

  /* either there is only 1 value or there are multiple values */
  if ( field->Count == 1 ) {
    if ( WriteSingleValue ( fp, field ) == ERROR ) {
      fprintf ( stderr, "cound not write data value ");
      fprintf ( stderr, "for tag number %d\n", field->Tag);
      return ( ERROR );
    }
  }
  else if ( field->Count > 1 ) {
    if ( WriteArrayOfValues ( fp, field ) == ERROR ) {
      fprintf ( stderr, "cound not write data values ");
      fprintf ( stderr, "for tag number %d\n", field->Tag);
      return ( ERROR );
    }
  }
  else {
    fprintf ( stderr, "tiff.c:  function WriteValue:\ncount for " );
    fprintf ( stderr, "field with tag %d is zero\n", field->Tag );
    return ( ERROR );
  }

  return ( NO_ERROR );
}

static int WriteArrayOfValues ( FILE *fp, struct TIFF_field *field )
{
  int i;

  if ( field->Type == BYTE ) {
    for ( i = 0; i < field->Count; i++ )
      if ( WriteUnsignedChar ( fp, &(field->Value.UCharArray[i]) )
           == ERROR ) return ( ERROR );
    return ( NO_ERROR );
  }
  else if ( field->Type == SHORT ) {
    for ( i = 0; i < field->Count; i++ )
      if ( WriteUnsignedShort ( fp, &(field->Value.UShortArray[i]) )
           == ERROR ) return ( ERROR );
    return ( NO_ERROR );
  }
  else if ( field->Type == LONG ) {
    for ( i = 0; i < field->Count; i++ )
      if ( WriteUnsignedLong ( fp, &(field->Value.ULongArray[i]) )
           == ERROR ) return ( ERROR );
    return ( NO_ERROR );
  }
  else if ( field->Type == RATIONAL ) {
    for ( i = 0; i < field->Count; i++ )
      if ( WriteRational ( fp, &(field->Value.RatArray[i]) )
           == ERROR ) return ( ERROR );
    return ( NO_ERROR );
  }

  fprintf ( stderr, "tiff.c:  function WriteArrayOfValues:\n" );
  fprintf ( stderr, "writer not prepared to write values of\n" );
  fprintf ( stderr, "type %d\n", field->Type );
  return ( ERROR );
}

static int WriteSingleValue ( FILE *fp, struct TIFF_field *field )
{
  if ( field->Type == BYTE ) {
    if ( WriteUnsignedChar ( fp, &(field->Value.UChar) ) == ERROR )
      return ( ERROR );
    return ( NO_ERROR );
  }
  else if ( field->Type == SHORT ) {
    if ( WriteUnsignedShort ( fp, &(field->Value.UShort) ) == ERROR )
      return ( ERROR );
    return ( NO_ERROR );
  }
  else if ( field->Type == LONG ) {
    if ( WriteUnsignedLong ( fp, &(field->Value.ULong) ) == ERROR )
      return ( ERROR );
    return ( NO_ERROR );
  }
  else if ( field->Type == RATIONAL ) {
    if ( WriteRational ( fp, &(field->Value.Rat) ) == ERROR )
      return ( ERROR );
    return ( NO_ERROR );
  }

  fprintf ( stderr, "tiff.c:  function WriteSingleValue:\n" );
  fprintf ( stderr, "writer not prepared to write value of\n" );
  fprintf ( stderr, "type %d\n", field->Type );
  return ( ERROR );
}

static int WriteHeader ( FILE *fp, struct TIFF_header *header )
{
  /* set file position at beginning of file */
  if ( SetFilePosition ( fp, 0 ) == ERROR ) {
    fprintf ( stderr, "tiff.c:  function WriteHeader:\n" );
    fprintf ( stderr, "error setting file position at byte 0\n" );
    return ( ERROR );
  }

  /* write byte-order word */
  if ( WriteUnsignedShort ( fp, &(header->ByteOrder) ) == ERROR )
    return ( ERROR );

  /* write forty-two */
  if ( WriteUnsignedShort ( fp, &(header->FortyTwo) ) == ERROR )
    return ( ERROR );

  /* write offset of first IFD */
  if ( WriteUnsignedLong ( fp, &(header->OffsetOfFirstIFD) ) 
       == ERROR )
    return ( ERROR );

  return ( NO_ERROR );
}

static int WriteRational ( FILE *fp, struct Rational *Rat )
{
  unsigned long numerator, denumerator;

  if (HOST_ORDER == TIFF_ORDER) {
    numerator   = Rat->Numer;
    denumerator = Rat->Denom;
  } else {
    numerator   = LongReverse(Rat->Numer);
    denumerator = LongReverse(Rat->Denom);
  }

  if ( ( 1 != fwrite ( &numerator,
                       sizeof ( unsigned long ), 1, fp ) ) ||
       ( 1 != fwrite ( &denumerator,
                       sizeof ( unsigned long ), 1, fp ) ) ) {
    fprintf ( stderr,"error writing rational number\n" );
    return ( ERROR );
  }

  return ( NO_ERROR );
}

static int WriteUnsignedLong ( FILE *fp, unsigned long *UnsignedLong )
{
  unsigned long unsignedlong;

  if (HOST_ORDER == TIFF_ORDER)
    unsignedlong = *UnsignedLong;
  else
    unsignedlong = LongReverse(*UnsignedLong);

  if ( 1 != fwrite ( ( unsigned long * ) &unsignedlong,
                     sizeof ( unsigned long ), 1, fp ) ) {
    fprintf ( stderr,"error writing unsigned long\n" );
    return ( ERROR );
  }

  return ( NO_ERROR );
}

static int WriteUnsignedShort ( FILE *fp, unsigned short *UnsignedShort )
{
  unsigned short unsignedshort;

  if (HOST_ORDER == TIFF_ORDER)
    unsignedshort = *UnsignedShort;
  else
    unsignedshort = ShortReverse(*UnsignedShort);

  if ( 1 != fwrite ( ( unsigned short * ) &unsignedshort,
                     sizeof ( unsigned short ), 1, fp ) ) {
    fprintf ( stderr,"error writing unsigned short\n" );
    return ( ERROR );
  }

  return ( NO_ERROR );
}

static int WriteUnsignedChar ( FILE *fp, unsigned char *UnsignedChar )
{
  if ( 1 != fwrite ( ( unsigned char * ) UnsignedChar,
                     sizeof ( unsigned char ), 1, fp ) ) {
    fprintf ( stderr,"error writing unsigned char\n" );
    return ( ERROR );
  }

  return ( NO_ERROR );
}

static int WriteImageData ( FILE *fp, struct TIFF_img *img,
                            struct DataLocation *DataLoc )
{
  /* write data uncompressed */
  if ( img->compress_type == 'u' )
    return ( WriteUncompressedImageData ( fp, img, DataLoc ) );

  fprintf ( stderr, "tiff.c:  function WriteImageData:\n" );
  fprintf ( stderr, "desired compression scheme " );
  fprintf ( stderr, "(compress_type '%c') not\nsupported by writer\n", 
            img->compress_type );
  return ( ERROR );
}

static int WriteUncompressedImageData ( FILE *fp, struct TIFF_img *img,
                                        struct DataLocation *DataLoc )
{
  unsigned long i, strip_offset;

  /* precompute rows per strip, bytes per row, and   */
  /* the number of strips; also allocate space for   */
  /* the arrays (strip_offsets and strip_byte_count) */
  /* in the DataLocation structure                   */
  if ( PrepDataLocStructUncompressed ( img, DataLoc ) == ERROR )
    return ( ERROR );

  /* set file position at 8th byte in file (right after header) */
  strip_offset = 8;
  if ( SetFilePosition ( fp, strip_offset ) == ERROR ) {
    fprintf ( stderr, "tiff.c:  function WriteUncompressedImage" );
    fprintf ( stderr, "Data:\nerror setting file position\n" );
    return ( ERROR );
  }

  /* write image data in strips */
  for ( i = 0; i < DataLoc->num_strips; i++ ) {

    /* record offset for strip */
    DataLoc->strip_offsets[i] = strip_offset;

    if ( WriteUncompressedStrip ( fp, img, DataLoc, i ) == ERROR ) 
      return ( ERROR );

    /* update strip_offset number */
    strip_offset += DataLoc->strip_byte_counts[i];

  }

  /* store the offset of the byte after the   */
  /* data (this may be the offset of the IFD, */
  /* but remember that the IFD has to begin   */
  /* on a word boundary - this is checked in  */
  /* PrepareHeader)                           */
  DataLoc->offset_of_byte_after_data = strip_offset;

  return ( NO_ERROR );
}

static int WriteUncompressedStrip ( FILE *fp, struct TIFF_img *img,
                                    struct DataLocation *DataLoc,
                                    unsigned long strip_index )
{
  unsigned long i, first_row_index;
 
  /* compute index of first row for this strip */
  first_row_index = strip_index * DataLoc->rows_per_strip;

  /* write one row at a time */
  for ( i = 0; i < DataLoc->rows_per_strip; i++ ) 
    if ( first_row_index + i < ( unsigned long ) ( img->height ) ) {
  
      if ( img->TIFF_type == 'g' ) {
        if ( WriteUncompGrayscaleRow ( fp, img, DataLoc, 
                                       first_row_index + i ) == ERROR ) 
          return ( ERROR );
      }
      else if ( img->TIFF_type == 'c' ) {
        if ( WriteUncompColorRow ( fp, img, DataLoc, 
                                   first_row_index + i ) == ERROR ) 
          return ( ERROR );
      }
      else {
        fprintf ( stderr, "tiff.c:  function WriteUncompressed" );
        fprintf ( stderr, "Strip:\nwriter is not prepared to " );
        fprintf ( stderr, "write data for\nimage of TIFF_type '%c'\n", 
                  img->TIFF_type );
        return ( ERROR );
      }
 
      DataLoc->strip_byte_counts[strip_index] += DataLoc->bytes_per_row;
  
    }

  return ( NO_ERROR );
}

static int WriteUncompGrayscaleRow ( FILE *fp, struct TIFF_img *img,
                                     struct DataLocation *DataLoc,
                                     unsigned long row_index )
{
  if ( DataLoc->bytes_per_row != 
       fwrite ( ( unsigned char * ) img->mono[row_index], 
                sizeof ( unsigned char ), 
                (size_t) DataLoc->bytes_per_row, fp ) ) 
    return ( ERROR );

  return ( NO_ERROR );
}

static int WriteUncompColorRow ( FILE *fp, struct TIFF_img *img,
                                 struct DataLocation *DataLoc,
                                 unsigned long row_index )
{
  int i, j;
  unsigned char *temp_row;

  /* allocate temporary space */
  if ( ( temp_row = ( unsigned char * ) 
                    calloc ( (size_t) DataLoc->bytes_per_row,   
                             sizeof ( unsigned char ) ) )
       == NULL ) {
    fprintf ( stderr, "tiff.c:  function WriteUncompColorRow:\n" );
    fprintf ( stderr, "couldn't allocate array of temp space\n" );
    return ( ERROR );
  }
     
  /* copy image data into space in RGBRGBRGB... order */
  for ( i = 0; i < img->width; i++ )
  for ( j = 0; j < 3; j++ )
    temp_row[ ( i * 3 ) + j ] = img->color[j][row_index][i];

  /* write row of image data */
  if ( DataLoc->bytes_per_row !=
       fwrite ( ( unsigned char * ) temp_row,
                sizeof ( unsigned char ),
                (size_t) DataLoc->bytes_per_row, fp ) )
    return ( ERROR );

  /* free up temp_row space */
  free ( ( void * ) temp_row );

  return ( NO_ERROR );
}

static int PrepDataLocStructUncompressed ( struct TIFF_img *img, 
                                           struct DataLocation *DataLoc )
{
  /* compute number of rows per strip so that there */
  /* are around 8000 bytes of data in each strip    */
  if ( ComputeRowsPerStrip ( img, DataLoc )
       == ERROR )
    return ( ERROR );

  /* compute number of strips necessary for    */
  /* storing image, store in DataLoc structure */
  if ( ComputeNumStrips ( img->height, DataLoc ) == ERROR )
    return ( ERROR );

  /* allocate arrays for DataLoc structure */
  if ( AllocateDataLocArrays ( DataLoc ) == ERROR )
    return ( ERROR );

  return ( NO_ERROR );
}

static int AllocateDataLocArrays ( struct DataLocation *DataLoc )
{
  if ( ( ( DataLoc->strip_offsets = 
             ( unsigned long * ) calloc ( (size_t) DataLoc->num_strips,
                                          sizeof ( unsigned long ) ) ) 
         == NULL ) || 
       ( ( DataLoc->strip_byte_counts = 
             ( unsigned long * ) calloc ( (size_t) DataLoc->num_strips,
                                          sizeof ( unsigned long ) ) )
         == NULL ) ) {
    fprintf ( stderr, "tiff.c:  function AllocateDataLocArrays:\n");
    fprintf ( stderr, "couldn't allocate arrays for DataLoc\n");
    return ( ERROR );
  }

  return ( NO_ERROR );
}

static int ComputeNumStrips ( int height, struct DataLocation *DataLoc )
{
  unsigned long ULheight;

  if ( height <= 0 ) {
    fprintf ( stderr, "tiff.c:  function ComputeNumStrips:\n" );
    fprintf ( stderr, "height entry in TIFF image structure\n" );
    fprintf ( stderr, "claims that image height is not positive\n" );
    return ( ERROR );
  }

  ULheight = ( unsigned long ) height;

  if ( ( ULheight % DataLoc->rows_per_strip ) == 0 )
    DataLoc->num_strips = ULheight / DataLoc->rows_per_strip;
  else 
    DataLoc->num_strips = 1 + ( unsigned long ) 
                              ( height / DataLoc->rows_per_strip );

  return ( NO_ERROR );
}

static int ComputeRowsPerStrip ( struct TIFF_img *img,
                                 struct DataLocation *DataLoc )
{
  unsigned long byte_count;

  /* uncompressed grayscale:  1 byte per pixel   */
  /* uncompressed full-color:  3 bytes per pixel */
  if ( img->TIFF_type == 'g' ) 
    DataLoc->bytes_per_row = img->width;
  else if ( img->TIFF_type == 'c' )  
         DataLoc->bytes_per_row = 3 * img->width;
  else {
    fprintf ( stderr, "tiff.c:  function ComputeBytesPerRow:\n" );
    fprintf ( stderr, "writer is not prepared to compute storage\n" );
    fprintf ( stderr, "info for image of TIFF_type '%c'\n", 
              img->TIFF_type );
    return ( ERROR );
  }

  /* compute rows_per_strip so that each strip will take up */
  /* about 8K of memory (cf. TIFF specification document);  */
  /* this number may be as low as 1 (for wide images) but   */
  /* not greater than the image height                      */
  DataLoc->rows_per_strip = 1;
  byte_count = DataLoc->bytes_per_row;
  while ( byte_count < 8000 ) {
    byte_count += DataLoc->bytes_per_row;
    DataLoc->rows_per_strip++;
  }

  if ( DataLoc->rows_per_strip > ( (unsigned long) img->height ) ) 
    DataLoc->rows_per_strip = (unsigned long) img->height;

  return ( NO_ERROR );
}

static int PrepareHeaderAndIFD ( struct TIFF_img *img, struct IFD *ifd,
                                 struct TIFF_header *header,
                                 struct DataLocation *DataLoc )
{
  /* prepare header */
  if ( PrepareHeader ( header, DataLoc ) == ERROR ) return ( ERROR );

  /* prepare IFD */
  if ( PrepareIFD ( img, ifd, DataLoc ) == ERROR ) return ( ERROR );

  return ( NO_ERROR );
}

static int PrepareHeader ( struct TIFF_header *header,
                           struct DataLocation *DataLoc )
{
/********************************/
/* This should be changed when  */
/* the code is modified so that */
/* it figures out byte-order by */
/* itself                       */
/********************************/
  /* put in byte-order word */
  header->ByteOrder = BigEndian;

  /* put in forty-two byte */
  header->FortyTwo = 42;

  /* compute offset to first byte of IFD; IFD */
  /* should be written right after image data */
  if ( DataLoc->offset_of_byte_after_data < 8 ) {
    fprintf ( stderr, "tiff.c:  function PrepareHeader:\ninvalid" );
    fprintf ( stderr, " entry for offset of IFD\n" );
    return ( ERROR ); 
  }
  header->OffsetOfFirstIFD = DataLoc->offset_of_byte_after_data;

  /* IFD should begin on a word boundary */
  if ( ( header->OffsetOfFirstIFD % 2 ) != 0 )
    header->OffsetOfFirstIFD++;

  return ( NO_ERROR );
}

static int PrepareIFD ( struct TIFF_img *img, 
                        struct IFD *ifd,
                        struct DataLocation *DataLoc )
{
  /* initialize IFD structure */
  ifd->NumberOfFields = 0;
  ifd->Fields = NULL;

  /* add field entries to IFD structure */
  if ( AddCoreFieldEntries ( img, ifd, DataLoc ) == ERROR ) 
    return ( ERROR );

  /* add fields which are required */
  /* for this specific image type  */
  if ( AddSpecialFieldEntries ( ifd, img->TIFF_type ) == ERROR ) 
    return ( ERROR );

  /* sort fields into ascending order of tag value */
  if ( SortFields ( ifd ) == ERROR ) return ( ERROR );

  return ( NO_ERROR );
}

static int AddSpecialFieldEntries ( struct IFD *ifd, char TIFF_type )
{
  /* for grayscale image */
  if ( TIFF_type == 'g' ) 
    return ( AddGrayscaleFields ( ifd, TIFF_type ) );

  /* for color image */
  if ( TIFF_type == 'c' ) 
    return ( AddColorFields ( ifd, TIFF_type ) );

  fprintf ( stderr, "tiff.c:  function AddSpecialFieldEntries:\n" );
  fprintf ( stderr, "writer is not prepared to add specialized\n" );
  fprintf ( stderr, "field entry for image of type %c\n", TIFF_type );
  return ( ERROR );
}

static int AddGrayscaleFields ( struct IFD *ifd, char TIFF_type )
{
  /* add MakeBitsPerSample field */
  if ( MakeBitsPerSampleField ( ifd, TIFF_type ) == ERROR ) 
    return ( ERROR );

  return ( NO_ERROR );
}

static int AddColorFields ( struct IFD *ifd, char TIFF_type )
{
  /* add MakeBitsPerSample field */
  if ( MakeBitsPerSampleField ( ifd, TIFF_type ) == ERROR ) 
    return ( ERROR );

  /* add SamplesPerPixel field */
  if ( MakeSamplesPerPixelField ( ifd ) == ERROR ) 
    return ( ERROR );

  return ( NO_ERROR );
}

static int SortFields ( struct IFD *ifd )
{
  unsigned long i, j, min_tag_index;

  /* for i = 0, ..., ifd->NOF-2, go through fields */
  /* j = i+1, ..., ifd->NOF - 1; find the j for    */
  /* which the tag value is lowest (among j's); if */
  /* this tag value is lower than tag of ith       */
  /* field, than switch the two fields             */

  for ( i = 0; i < ifd->NumberOfFields - 1; i++ ) {

    min_tag_index = i;

    for ( j = i + 1; j < ifd->NumberOfFields; j++ )
      if ( ifd->Fields[j].Tag < ifd->Fields[min_tag_index].Tag )
        min_tag_index = j;

    if ( min_tag_index != i ) 
      SwitchFields ( &(ifd->Fields[i]), 
                     &(ifd->Fields[min_tag_index]) );

  }

  return ( NO_ERROR );
}

static void SwitchFields ( struct TIFF_field *FieldOne, 
                           struct TIFF_field *FieldTwo ) 
{
  struct TIFF_field temp;

  temp = *FieldOne;
  *FieldOne = *FieldTwo;
  *FieldTwo = temp;
}

static int AddCoreFieldEntries ( struct TIFF_img *img,
                                 struct IFD *ifd,
                                 struct DataLocation *DataLoc )
{
  /* add ImageWidth field */
  if ( MakeImageWidthOrLengthField ( ifd, ImageWidth, img->width ) 
       == ERROR ) return ( ERROR );

  /* add ImageLength field */
  if ( MakeImageWidthOrLengthField ( ifd, ImageLength, img->height ) 
       == ERROR ) return ( ERROR );

  /* add Compression field */
  if ( MakeCompressionField ( ifd, img->compress_type ) == ERROR ) 
    return ( ERROR );

  /* add PhotometricInterpretation field */
  if ( MakePhotometricInterpretationField ( ifd, img->TIFF_type ) 
       == ERROR ) return ( ERROR );

  /* add StripOffsets field */
  if ( MakeStripOffsetsField ( ifd, DataLoc ) == ERROR )
    return ( ERROR );

  /* add RowsPerStrip field */
  if ( MakeRowsPerStripField ( ifd, DataLoc ) == ERROR )
    return ( ERROR );

  /* add StripByteCounts field */
  if ( MakeStripByteCountsField ( ifd, DataLoc ) == ERROR )
    return ( ERROR );

  /* add XResolution field */
  if ( MakeResolutionField ( ifd, XResolution ) == ERROR )
    return ( ERROR );

  /* add YResolution field */
  if ( MakeResolutionField ( ifd, YResolution ) == ERROR )
    return ( ERROR );

  /* add ResolutionUnit field */
  if ( MakeResolutionUnitField ( ifd ) == ERROR ) return ( ERROR );

  return ( NO_ERROR );
}

static int MakeSamplesPerPixelField ( struct IFD *ifd )
{
  struct TIFF_field *field;

  /* increment NumberOfFields in IFD struct */
  ifd->NumberOfFields++;

  /* allocate field structure */
  if ( AllocateNewField ( ifd ) == ERROR ) return ( ERROR );
  field = &(ifd->Fields[ifd->NumberOfFields - 1]);

  field->Tag = SamplesPerPixel;
  field->Type = SHORT;
  field->Count = 1;
  field->Value.UShort = 3;

  return ( NO_ERROR );
}

static int MakeBitsPerSampleField ( struct IFD *ifd, char TIFF_type )
{
  struct TIFF_field *field;

  /* increment NumberOfFields in IFD struct */
  ifd->NumberOfFields++;

  /* allocate field structure */
  if ( AllocateNewField ( ifd ) == ERROR ) return ( ERROR );
  field = &(ifd->Fields[ifd->NumberOfFields - 1]);

  field->Tag = BitsPerSample;
  field->Type = SHORT;
 
  /* if image is grayscale */
  if ( TIFF_type == 'g' ) {
    field->Count = 1;
    field->Value.UShort = 8;
    return ( NO_ERROR );
  }

  /* if image is color */
  if ( TIFF_type == 'c' ) {
    field->Count = 3;

    /* allocate array of values */
    if ( AllocateArrayOfValues ( field ) == ERROR ) {
      fprintf ( stderr, "tiff.c:  function MakeBitsPerSample" );
      fprintf ( stderr, "Field:\ncouldn't alloc values for field\n" );
      return ( ERROR );
    }

    field->Value.UShortArray[0] = 8;
    field->Value.UShortArray[1] = 8;
    field->Value.UShortArray[2] = 8;
    return ( NO_ERROR );
  }

  fprintf ( stderr, "tiff.c:  function MakeBitsPerSampleField:\n" );
  fprintf ( stderr, "writer not prepared to write field value(s)\n" );
  fprintf ( stderr, "for BitsPerSample field for image of type %c\n", 
            TIFF_type );
  return ( ERROR );
}

static int MakeResolutionUnitField ( struct IFD *ifd )
{
  struct TIFF_field *field;

  /* increment NumberOfFields in IFD struct */
  ifd->NumberOfFields++;
 
  /* allocate field structure */
  if ( AllocateNewField ( ifd ) == ERROR ) return ( ERROR );
  field = &(ifd->Fields[ifd->NumberOfFields - 1]);

  /* put in default values */
  field->Tag = ResolutionUnit;
  field->Type = SHORT;
  field->Count = 1;
  field->SizeOfType = 2;
  field->Value.UShort = 2;

  return ( NO_ERROR );
}

static int MakeResolutionField ( struct IFD *ifd, unsigned short Tag ) 
{
  struct TIFF_field *field;

  /* increment NumberOfFields in IFD struct */
  ifd->NumberOfFields++;

  /* allocate field structure */
  if ( AllocateNewField ( ifd ) == ERROR ) return ( ERROR );
  field = &(ifd->Fields[ifd->NumberOfFields - 1]);

  /* assign a default resolution of 1200 dpi */
  /* (this was the res assigned for a TIFF   */
  /* file written out by XV)                 */
  field->Tag = Tag;
  field->Type = RATIONAL;
  field->Count = 1;
  field->Value.Rat.Numer = 314572800;
  field->Value.Rat.Denom = 262144;

  return ( NO_ERROR );
}

static int MakeStripByteCountsField ( struct IFD *ifd,
                                      struct DataLocation *DataLoc )
{
  struct TIFF_field *field;
  unsigned long max_byte_count, i;

  /* increment NumberOfFields in IFD struct */
  ifd->NumberOfFields++;

  /* allocate field structure */
  if ( AllocateNewField ( ifd ) == ERROR ) return ( ERROR );
  field = &(ifd->Fields[ifd->NumberOfFields - 1]);

  field->Tag = StripByteCounts;
  field->Count = DataLoc->num_strips;

  max_byte_count = GetMaxValUL ( DataLoc->strip_byte_counts, 
                                 DataLoc->num_strips );

  /* number of strip offsets should fit into LONG or SHORT */
  if ( max_byte_count < 65536 ) {
    field->Type = SHORT;

    if ( field->Count == 1 )
      field->Value.UShort = ( unsigned short ) 
                            DataLoc->strip_byte_counts[0];
    else {
      if ( AllocateArrayOfValues ( field ) == ERROR ) return ( ERROR );
      for ( i = 0; i < field->Count; i++ )
        field->Value.UShortArray[i] =
          ( unsigned short ) DataLoc->strip_byte_counts[i];
    }
  }
  else {
    field->Type = LONG;

    if ( field->Count == 1 ) 
      field->Value.ULong = DataLoc->strip_byte_counts[0];
    else {
      if ( AllocateArrayOfValues ( field ) == ERROR ) return ( ERROR );
      for ( i = 0; i < field->Count; i++ )
        field->Value.ULongArray[i] = DataLoc->strip_byte_counts[i];
    }
  }

  return ( NO_ERROR );
}

static int MakeRowsPerStripField ( struct IFD *ifd,
                                   struct DataLocation *DataLoc )
{
  struct TIFF_field *field;

  /* increment NumberOfFields in IFD struct */
  ifd->NumberOfFields++;

  /* allocate field structure */
  if ( AllocateNewField ( ifd ) == ERROR ) return ( ERROR );
  field = &(ifd->Fields[ifd->NumberOfFields - 1]);

  field->Tag = RowsPerStrip;
  field->Count = 1;

  /* number of RowsPerStrip should fit into LONG or SHORT */
  if ( DataLoc->rows_per_strip == 0 ) {
    fprintf ( stderr, "tiff.c:  function MakeRowsPerStripField:\n" );
    fprintf ( stderr, "element in DataLocation structure claims\n" );
    fprintf ( stderr, "that the number of RowsPerStrip is zero\n" );
    return ( ERROR );
  }
  if ( DataLoc->rows_per_strip < 65536 ) {
    field->Type = SHORT;
    field->Value.UShort = ( unsigned short ) DataLoc->rows_per_strip;
  }
  else {
    field->Type = LONG;
    field->Value.ULong = DataLoc->rows_per_strip;
  }

  return ( NO_ERROR );
}

static int MakeStripOffsetsField ( struct IFD *ifd, 
                                   struct DataLocation *DataLoc )
{
  struct TIFF_field *field;
  unsigned long i;

  /* increment NumberOfFields in IFD struct */
  ifd->NumberOfFields++;

  /* allocate field structure */
  if ( AllocateNewField ( ifd ) == ERROR ) return ( ERROR );
  field = &(ifd->Fields[ifd->NumberOfFields - 1]);

  field->Tag = StripOffsets;
  field->Type = LONG;
  field->Count = DataLoc->num_strips;

  /* number of strip offsets should not be zero */
  if ( field->Count == 1 ) 
    field->Value.ULong = DataLoc->strip_offsets[0];
  else if ( field->Count > 1 ) {
    if ( AllocateArrayOfValues ( field ) == ERROR ) return ( ERROR );
    for ( i= 0; i < field->Count; i++ )
      field->Value.ULongArray[i] = DataLoc->strip_offsets[i];
  }
  else {
    fprintf ( stderr, "tiff.c:  function MakeStripOffsets" );
    fprintf ( stderr, "Field:\nelement in DataLocation structure " );
    fprintf ( stderr, "claims that\nthe number of strips " );
    fprintf ( stderr, "which were written out is zero\n" );
    return ( ERROR );
  }

  return ( NO_ERROR );
}

static int MakePhotometricInterpretationField ( struct IFD *ifd, char TIFF_type )
{
  struct TIFF_field *field;

  /* increment NumberOfFields in IFD struct */
  ifd->NumberOfFields++;

  /* allocate field structure */
  if ( AllocateNewField ( ifd ) == ERROR ) return ( ERROR );
  field = &(ifd->Fields[ifd->NumberOfFields - 1]);

  field->Tag = PhotometricInterpretation;
  field->Type = SHORT;
  field->Count = 1;

  if ( TIFF_type == 'g' ) field->Value.UShort = BlackIsZero;
  else if ( TIFF_type == 'c' ) field->Value.UShort = RGB;
  else {
    fprintf ( stderr, "tiff.c:  function MakePhotometricInterpretation" );
    fprintf ( stderr, "Field:\ndesired image type " );
    fprintf ( stderr, "(TIFF_type '%c') not supported by writer\n",
              TIFF_type );
    return ( ERROR );
  }

  return ( NO_ERROR );
}

static int MakeCompressionField ( struct IFD *ifd, char compress_type )
{
  struct TIFF_field *field;

  /* increment NumberOfFields in IFD struct */
  ifd->NumberOfFields++;

  /* allocate field structure */
  if ( AllocateNewField ( ifd ) == ERROR ) return ( ERROR );
  field = &(ifd->Fields[ifd->NumberOfFields - 1]);

  field->Tag = Compression;
  field->Type = SHORT;
  field->Count = 1;

  if ( compress_type == 'u' ) field->Value.UShort = NoCompression;
  else {
    fprintf ( stderr, "tiff.c:  function MakeCompressionField:\n" );
    fprintf ( stderr, "desired compression scheme " );
    fprintf ( stderr, "(compress_type '%c') not supported by writer\n",
              compress_type );
    return ( ERROR );
  }

  return ( NO_ERROR );
}

static int MakeImageWidthOrLengthField ( struct IFD *ifd, 
                                         unsigned short Tag,
                                         int dimension )
{
  struct TIFF_field *field;

  /* increment NumberOfFields in IFD struct */
  ifd->NumberOfFields++;

  /* allocate field structure */
  if ( AllocateNewField ( ifd ) == ERROR ) return ( ERROR );
  field = &(ifd->Fields[ifd->NumberOfFields - 1]);


  field->Tag = Tag;
  field->Count = 1;

  /* dimension should not be too low */
  if ( dimension < 1 ) {
    fprintf ( stderr, "tiff.c:  function MakeImageWidthOrLength" );
    fprintf ( stderr, "Field:\nimage dimension is less than 1\n" );
    return ( ERROR );
  }

  /* dimension should fit into a SHORT or a LONG */
  if ( dimension < 65536 ) {
    field->Type = SHORT;
    field->Value.UShort = ( unsigned short ) dimension;
  }
  else {
    field->Type = LONG;
    field->Value.ULong = ( unsigned long ) dimension;
  }

  return ( NO_ERROR );
}

static unsigned long GetMaxValUL ( unsigned long *array, 
                                   unsigned long n_elements )
{
  unsigned long i, max_index;

  max_index = 0;
  for ( i = 1; i < n_elements; i++ )
    if ( array[max_index] < array[i] )
      max_index = i;

  return ( array[max_index] );
}

int read_TIFF ( FILE *fp, struct TIFF_img *img )
{
  struct IFD ifd;
  struct TIFF_header header;

  if ((HOST_ORDER = tiffHostType(0)) == UnknownEndian) return ( ERROR );

  /* ensure that each type will be stored */
  /* in the expected number of bytes      */
  if ( CheckTypeSizes ( ) == ERROR ) return ( ERROR );

  /* read header */
  if ( ReadHeader ( fp, &(header) ) == ERROR ) return ( ERROR );

  /* read image file directory */
  if ( ReadIFD ( fp, &(ifd), &(header) ) == ERROR ) 
    return ( ERROR );

  /* verify that IFD is complete; fill in for    */
  /* missing fields which have default value(s); */
  /* determine and record "TIFF_type" of image   */
  if ( VerifyIFD ( &(ifd), &(img->TIFF_type) ) == ERROR ) 
    return ( ERROR );

  /* read image data */
  if ( GetImageData ( fp, img, &(ifd) ) == ERROR ) 
    return ( ERROR );

  return ( NO_ERROR );
}

int read_TIFF_header ( FILE *fp, struct TIFF_img *img )
{
  struct IFD ifd;
  struct TIFF_header header;

  if ((HOST_ORDER = tiffHostType(0)) == UnknownEndian) return ( ERROR );

  /* ensure that each type will be stored */
  /* in the expected number of bytes      */
  if ( CheckTypeSizes ( ) == ERROR ) return ( ERROR );

  /* read header */
  if ( ReadHeader ( fp, &(header) ) == ERROR ) return ( ERROR );

  /* read image file directory */
  if ( ReadIFD ( fp, &(ifd), &(header) ) == ERROR ) 
    return ( ERROR );

  /* verify that IFD is complete; fill in for    */
  /* missing fields which have default value(s); */
  /* determine and record "TIFF_type" of image   */
  if ( VerifyIFD ( &(ifd), &(img->TIFF_type) ) == ERROR ) 
    return ( ERROR );

  return ( NO_ERROR );
}
static int GetImageData ( FILE *fp, struct TIFF_img *img, struct IFD *ifd )
{
  /* ensure compression scheme is recognized; */
  /* if it is, record compress_type           */
  if ( GetCompression ( ifd, &(img->compress_type) ) == ERROR ) 
    return ( ERROR );

  /* record image height and width */
  if ( GetHeightAndWidth ( ifd, &(img->height), &(img->width) ) 
       == ERROR ) return ( ERROR );

  /* allocate array for image data */
  if ( AllocateImageDataArray ( img ) == ERROR ) return ( ERROR );

  /* copy image data into the array */
  if ( CopyImageData ( fp, img, ifd ) == ERROR ) return ( ERROR );

  return ( NO_ERROR );
}

static int CopyImageData ( FILE *fp, struct TIFF_img *img, struct IFD *ifd )
{
  struct DataLocation DataLoc;

  /* get information about location of image data */
  if ( GetImageDataLocationInfo ( ifd, &(DataLoc) ) == ERROR ) 
    return ( ERROR );

  /* if data is uncompressed */
  if ( img->compress_type == 'u' ) {
    if ( ( CopyUncompressedImageData ( fp, img, ifd, &(DataLoc) ) ) 
         == ERROR ) return ( ERROR );
  }
  else {
    fprintf ( stderr, "tiff.c:  function CopyImageData:\n" );
    fprintf ( stderr, "reader is not prepared to decompress data\n" );
    fprintf ( stderr, "compressed with compress_type '%c'\n",
              img->compress_type );
    return ( ERROR );
  }
 
  /* free arrays in DataLocation structure */ 
  FreeDataLocation ( &DataLoc );

  return ( NO_ERROR );
}

static void FreeDataLocation ( struct DataLocation *DataLoc )
{
  free ( ( void * ) DataLoc->strip_offsets );
  free ( ( void * ) DataLoc->strip_byte_counts );
}

static int CopyUncompressedImageData ( FILE *fp, struct TIFF_img *img, 
                                       struct IFD *ifd,
                                       struct DataLocation *DataLoc )
{
  unsigned long i;

  /* loop through strips */
  for ( i = 0; i < DataLoc->num_strips; i++ ) {

    /* set file-position at first byte in strip */
    if ( SetFilePosition ( fp, DataLoc->strip_offsets[i] ) == ERROR ) {
      fprintf ( stderr, "error finding strip %d of image data\n",
                DataLoc->strip_offsets[i] );
      return ( ERROR );
    }

    /* get strip of image data */
    if ( GetUncompressedStrip ( fp, img, ifd, DataLoc, i ) 
         == ERROR ) 
      return ( ERROR );

  }

  return ( NO_ERROR );
}

static int GetUncompressedStrip ( FILE *fp, struct TIFF_img *img, 
                                  struct IFD *ifd,
                                  struct DataLocation *DataLoc,
                                  unsigned long strips_already_read )
{
  unsigned long i, current_row;
  
  /* read image data one row at a time */
  for ( i = 0; i < DataLoc->rows_per_strip; i++ )
    if ( IsThereAnyMoreDataInThisStrip ( ifd, img->width, DataLoc, 
                                         strips_already_read, i ) 
         == YES ) {

      /* get another row of data */
      current_row = (DataLoc->rows_per_strip * strips_already_read) + i;
      if ( img->TIFF_type == 'g' ) {
        if ( GetUncompressedGrayscaleRow ( fp, img, current_row )
             == ERROR )
          return ( ERROR );
      }
      else if ( img->TIFF_type == 'c' ) {
          if ( GetUncompressedColorRow ( fp, img, current_row ) == ERROR )
            return ( ERROR );
      }
      else {
        fprintf ( stderr, "tiff.c:  function GetUncompressedStrip\n" );
        fprintf ( stderr, "can't get row of data from image of\n" );
        fprintf ( stderr, "type %c\n", img->TIFF_type );
        return ( ERROR );
      }

    }

  return ( NO_ERROR );
}

static int GetUncompressedColorRow ( FILE *fp, struct TIFF_img *img,
                                     unsigned long current_row )
{
  unsigned long i, j;
  
  /* assume data is stored in "chunky" format; */
  /* i.e., component values for each pixel are */
  /* are stored contiguously: RGBRGBRGB...;    */
  /* this is the default planar configuration  */
  for ( i = 0; i < img->width; i++ ) 
    for ( j = 0; j < 3; j++ )
      if ( 1 != fread ( &(img->color[j][current_row][i]),
                        sizeof ( unsigned char ), 1, fp ) ) {
        fprintf ( stderr, "tiff.c:  function GetUncompressedColor" );
        fprintf ( stderr, "Row:\nerror reading image data\n" );
        return ( ERROR );
      }

  return ( NO_ERROR );
}

static int GetUncompressedGrayscaleRow ( FILE *fp, struct TIFF_img *img,
                                         unsigned long current_row )
{
  if ( img->width != fread ( ( unsigned char * ) img->mono[current_row],
                             sizeof ( unsigned char ),
                             ( size_t ) img->width, fp ) ) {
    fprintf ( stderr, "tiff.c:  function GetUncompressedGrayscale" );
    fprintf ( stderr, "Row:\nerror reading image data\n" );
    return ( ERROR );
  }

  return ( NO_ERROR );
}

static int IsThereAnyMoreDataInThisStrip ( struct IFD *ifd, int width,
                                           struct DataLocation *DataLoc,
                                           unsigned long current_strip_index,
                                           unsigned long rows_already_read )
{
  unsigned long BytesPerRow, BytesAlreadyRead;
 
  /* compute number of bytes of file that */
  /* are used to store one row of pixels  */
  if ( ComputeBytesPerRow ( ifd, width, &(BytesPerRow) ) == ERROR ) 
    return ( ERROR );

  /* compute number of bytes already taken from this strip */
  BytesAlreadyRead = rows_already_read * BytesPerRow;

  if ( BytesAlreadyRead <
       DataLoc->strip_byte_counts[current_strip_index] )
    return ( YES );

  return ( NO );
}

static int ComputeBytesPerRow ( struct IFD *ifd, int width,
                                unsigned long *BytesPerRow )
{
  unsigned long BytesPerSample;

  /* compute bytes per pixel in image */
  if ( ComputeBytesPerSample ( ifd, &(BytesPerSample) ) == ERROR ) 
    return ( ERROR );

  /* number of BytesPerRow is number of BytesPerSample  */
  /* times the number of pixels in a row (image width) */
  *BytesPerRow = BytesPerSample * ( unsigned long ) ( width );

  return ( NO_ERROR );
}

static int ComputeBytesPerSample ( struct IFD *ifd, 
                                   unsigned long *BytesPerSample )
{
  unsigned short samples_per_pixel;
  unsigned long total_bits_per_sample;

  /* samples_per_pixel is usually 1 for bilevel, */
  /* grayscale, and palette-color images         */
  samples_per_pixel = 1;

  /* for full-color images this may be more than 3 */
  if ( IsThereAFieldFor ( ifd, SamplesPerPixel ) == YES )
    if ( GetSamplesPerPixel ( ifd, &(samples_per_pixel) ) == ERROR ) 
      return ( ERROR );

  if ( ( samples_per_pixel == 2 ) || 
       ( samples_per_pixel > 3 ) ) {
    fprintf ( stderr, "tiff.c:  function ComputeBytesPerSample:\n" );
    fprintf ( stderr, "value (%d) for SamplesPerPixel field\n", 
              samples_per_pixel );
    fprintf ( stderr, "suggests a situation for which tiff reader\n" );
    fprintf ( stderr, "is not prepared\n" );
    return ( ERROR );
  }

  /* compute total bits per sample */
  if ( ComputeTotalBitsPerSample ( ifd, &(total_bits_per_sample) )
       == ERROR ) return ( ERROR );

  if ( ( total_bits_per_sample % 8 ) != 0 ) {
    fprintf ( stderr, "tiff.c:  function ComputeBytesPerSample:\n" );
    fprintf ( stderr, "total_bits_per_sample is not a multiple\n" );
    fprintf ( stderr, "of eight, which suggests a situation for\n" );
    fprintf ( stderr, "which tiff reader is not prepared\n" );
    return ( ERROR );
  }

  *BytesPerSample = total_bits_per_sample / 8;

  return ( NO_ERROR );
}

static int ComputeTotalBitsPerSample ( struct IFD *ifd, unsigned long *Total )
{
  unsigned long i;
  struct TIFF_field *field;

  /* get pointer to BitsPerSample field */
  if ( IsThereAFieldFor ( ifd, BitsPerSample ) == YES ) {
    if ( ( field = GetFieldStructure ( ifd, BitsPerSample ) ) == NULL )
      return ( ERROR );
  }
  else {
    fprintf ( stderr, "tiff.c:  function ComputeTotalBitsPerSample\n" );
    fprintf ( stderr, "field for BitsPerSample is not present\n" );
    return ( ERROR );
  }

  if ( field->Type == SHORT ) {
    if ( field->Count == 1 ) *Total = field->Value.UShort;
    if ( field->Count > 1 ) {
      *Total = 0;
      for ( i = 0; i < field->Count; i++ )
        *Total += ( unsigned long ) ( field->Value.UShortArray[i] );
    }
  }
  else {
    WrongValueType ( "ComputeTotalBitsPerSample",
                     field->Tag, field->Type );
    return ( ERROR );
  }

  return ( NO_ERROR ); 
}

static int GetSamplesPerPixel ( struct IFD *ifd, unsigned short *Number )
{
  struct TIFF_field *field;

  /* get pointer to SamplesPerPixel field */
  if ( ( field = GetFieldStructure ( ifd, SamplesPerPixel ) ) == NULL )
    return ( ERROR );

  if ( field->Type == SHORT ) *Number = field->Value.UShort;
  else {
    WrongValueType ( "GetSamplesPerPixel", field->Tag, field->Type );
    return ( ERROR );
  }

  return ( NO_ERROR );
}

static int GetCompressionValue ( struct IFD *ifd, unsigned short *Value )
{
  struct TIFF_field *field;

  /* get pointer to Compression field */
  if ( ( field = GetFieldStructure ( ifd, Compression ) ) == NULL )
    return ( ERROR );

  if ( field->Type == SHORT ) *Value = field->Value.UShort;
  else {
    WrongValueType ( "GetCompressionValue", field->Tag, field->Type );
    return ( ERROR );
  }

  return ( NO_ERROR );
}

static int GetImageDataLocationInfo ( struct IFD *ifd, 
                                      struct DataLocation *DataLoc )
{
  /* get RowsPerStrip value */
  if ( GetRowsPerStrip ( ifd, &(DataLoc->rows_per_strip) ) == ERROR )
    return ( ERROR );

  /* get number of strips */
  if ( GetNumberOfStrips ( ifd, &(DataLoc->num_strips) ) == ERROR )
    return ( ERROR );

  /* get array of strip offsets */
  if ( GetStripOffsets ( ifd, DataLoc ) == ERROR )
    return ( ERROR );

  /* get array of strip byte counts */
  if ( GetStripByteCounts ( ifd, DataLoc ) == ERROR )
    return ( ERROR );

  return ( NO_ERROR );
}

static int GetStripOffsets ( struct IFD *ifd,
                             struct DataLocation *DataLoc )
{
  struct TIFF_field *field;

  /* get pointer to StripOffsets field */
  if ( ( field = GetFieldStructure ( ifd, StripOffsets ) )
       == NULL ) return ( ERROR );

  /* stuff field values into array in DataLocation structure */
  if ( PutValuesIntoULongArray ( field, &(DataLoc->strip_offsets), 
                                 DataLoc->num_strips ) 
       == ERROR )
    return ( ERROR );

  return ( NO_ERROR );
}

static int GetStripByteCounts ( struct IFD *ifd,
                                struct DataLocation *DataLoc )
{
  struct TIFF_field *field;

  /* get pointer to StripByteCounts field */
  if ( ( field = GetFieldStructure ( ifd, StripByteCounts ) )
       == NULL ) return ( ERROR );

  /* stuff field values into array in DataLocation structure */
  if ( PutValuesIntoULongArray ( field, &(DataLoc->strip_byte_counts), 
                                 DataLoc->num_strips ) 
       == ERROR )
    return ( ERROR );

  return ( NO_ERROR );
}

static int PutValuesIntoULongArray ( struct TIFF_field *field,
                                     unsigned long **array,
                                     unsigned long num_elements )
{
  int unsigned long i;

  /* allocate array */
  if ( ( *array = ( unsigned long * ) 
                  calloc ( ( size_t ) num_elements,
                           sizeof ( unsigned long ) ) )
       == NULL ) {
    fprintf ( stderr, "tiff.c:  function PutValuesInto" );
    fprintf ( stderr, "ULongArray:\n" );
    fprintf ( stderr, "couldn't allocate array of unsigned long\n");
    return ( ERROR );
  }

  /* ensure array of field values has expected numer of elements */
  if ( field->Count != num_elements ) {
    fprintf ( stderr, "tiff.c:  function PutValuesIntoULongArray:\n" );
    fprintf ( stderr, "field count (%d) for tag %d ", 
              field->Count, field->Tag ); 
    fprintf ( stderr, "differs from num_strips element (%d)\n" );
    fprintf ( stderr, "in DataLocation structure\n" );
    return ( ERROR );
  }

  /* stuff field values into array */
  if ( field->Type == SHORT ) {
    if ( field->Count == 1 ) 
      (*array)[0] = ( unsigned long ) field->Value.UShort;
    else {
      for ( i = 0; i < num_elements; i++ ) 
        (*array)[i] = ( unsigned long ) ( field->Value.UShortArray[i] );
    }
  }
  else if ( field->Type == LONG ) {
    if ( field->Count == 1 ) (*array)[0] = field->Value.ULong;
    else {
      for ( i = 0; i < num_elements; i++ )
        (*array)[i] = field->Value.ULongArray[i];
    }
  }
  else {
    WrongValueType ( "PutValuesIntoULongArray",
                     field->Tag, field->Type );
    return ( ERROR );
  }

  return ( NO_ERROR );
}

static int GetNumberOfStrips ( struct IFD *ifd,
                               unsigned long *num_strips )
{
  struct TIFF_field *field;

  /* get pointer to StripOffsets field */
  if ( ( field = GetFieldStructure ( ifd, StripOffsets ) )
       == NULL ) return ( ERROR );

  /* get number of strips from field Count */
  *num_strips = field->Count;

  return ( NO_ERROR );
}

static int GetRowsPerStrip ( struct IFD *ifd, 
                             unsigned long *rows_per_strip )
{
  struct TIFF_field *field;

  /* get pointer to RowsPerStrip field */
  if ( ( field = GetFieldStructure ( ifd, RowsPerStrip ) )
       == NULL ) return ( ERROR );

  /* get number of rows per strip */
  if ( field->Type == SHORT )
    *rows_per_strip = ( unsigned long ) field->Value.UShort;
  else if ( field->Type == LONG ) *rows_per_strip = field->Value.ULong;
  else {
    WrongValueType ( "GetRowsPerStrip", field->Tag, field->Type );
    return ( ERROR );
  }

  return ( NO_ERROR );
}

static void WrongValueType ( char *FunctionName, unsigned short Tag,
                             unsigned short Type )
{
    fprintf ( stderr, "tiff.c:  function %s:\n", FunctionName );
    fprintf ( stderr, "tiff reader is not prepared for interpreting\n" );
    fprintf ( stderr, "value type %d for field with tag %d\n", Type, Tag );
}

static int AllocateImageDataArray ( struct TIFF_img *img )
{
  int i;

  /* if image is grayscale allocate space for "mono" array */
  if ( img->TIFF_type == 'g' ) {
    img->mono = ( unsigned char ** ) get_img ( img->width, img->height,
                                               sizeof ( unsigned char ) );
    return ( NO_ERROR );
  }

  /* if image is full color allocate */
  /* one plane each for R, G, B      */
  if ( img->TIFF_type == 'c' ) {
    img->color = ( unsigned char *** ) 
                 mget_spc ( 3, sizeof ( unsigned char ** ) );
    for ( i = 0; i < 3; i++ )
      img->color[i] = ( unsigned char ** ) 
                      get_img ( img->width, img->height,
                                sizeof ( unsigned char ) );
    return ( NO_ERROR );
  }

  fprintf ( stderr, "tiff.c:  function AllocateImageDataArray:\n" );
  fprintf ( stderr, "space allocation for image of type %c ", 
            img->TIFF_type );
  fprintf ( stderr, "is not supported\n" );
  return ( ERROR );
}

static int GetHeightAndWidth ( struct IFD *ifd, int *height, int *width )
{
  struct TIFF_field *field;

  /* get pointer to ImageLength field */
  if ( ( field = GetFieldStructure ( ifd, ImageLength ) )
       == NULL ) return ( ERROR );

  /* get image height */
  if ( field->Type == SHORT ) *height = ( int ) field->Value.UShort;
  else if ( field->Type == LONG ) *height = ( int ) field->Value.ULong;
  else {
    WrongValueType ( "GetHeightAndWidth", field->Tag, field->Type );
    return ( ERROR );
  }

  /* get pointer to ImageWidth field */
  if ( ( field = GetFieldStructure ( ifd, ImageWidth ) )
       == NULL ) return ( ERROR );

  /* get image width */
  if ( field->Type == SHORT ) *width = ( int ) field->Value.UShort;
  else if ( field->Type == LONG ) *width = ( int ) field->Value.ULong;
  else {
    WrongValueType ( "GetHeightAndWidth", field->Tag, field->Type );
    return ( ERROR );
  }

  return ( NO_ERROR );
}

static int GetCompression ( struct IFD *ifd, char *compress_type )
{
  unsigned short CompressionValue;

  /* get compression value */
  if ( GetCompressionValue ( ifd, &CompressionValue ) == ERROR )
    return ( ERROR );

  if ( CompressionValue == NoCompression ) {
    *compress_type = 'u';
    return ( NO_ERROR );
  }

  if ( CompressionValue == PackBits ) {
    fprintf ( stderr, "tiff.c:  function GetCompression:\n" );
    fprintf ( stderr, "PackBits compression not supported\n" );
    return ( ERROR );
  }

  if ( CompressionValue == OneDimModifiedHuffman ) {
    fprintf ( stderr, "tiff.c:  function GetCompression:\n" );
    fprintf ( stderr, "1-D modified Huffman compression not supported\n" );
    return ( ERROR );
  }

  fprintf ( stderr, "tiff.c:  function GetCompression:\n" );
  fprintf ( stderr, "Compression value %d not recognized; could not", 
            CompressionValue );
  fprintf ( stderr, " assign\ncompress_type to TIFF_img structure\n" );
  return ( ERROR );
}

static int VerifyIFD ( struct IFD *ifd, char *TIFF_type )
{
  /* ensure that there are IFD entries for all */
  /* "core" fields (those which are required   */
  /* for each of the baseline image types); if */
  /* any of these is missing and has a default */
  /* value, add the field (with default value) */
  /* to the IFD                                */
  if ( CheckForCoreFields ( ifd ) == ERROR ) return ( ERROR ); 

  /* fill in TIFF_img structure element for  */
  /* image type (this involves considerable  */
  /* error checking, which is why it is part */
  /* of the IFD verification procedure)      */
  if ( GetImageType ( ifd, TIFF_type ) == ERROR ) return ( ERROR );
  
  return ( NO_ERROR );
}

static int GetImageType ( struct IFD *ifd, char *TIFF_type )
{
  /* grayscale */
  if ( IsImageGrayscale ( ifd ) == YES ) {
    *TIFF_type = 'g';
    return ( NO_ERROR );
  }

  /* full 24-bit color */
  if ( IsImageFullColor ( ifd ) == YES ) {
    *TIFF_type = 'c';
    return ( NO_ERROR );
  }

  fprintf ( stderr, "tiff.c:  function GetImageType:\nreader " );
  fprintf ( stderr, "can't infer image type from IFD fields" );
  return ( ERROR );
}

static int IsImageFullColor ( struct IFD *ifd )
{
  struct TIFF_field *field;

  /* there must be fields for          */
  /* BitsPerSample and SamplesPerPixel */
  if ( IsThereAFieldFor ( ifd, BitsPerSample ) == NO ) return ( NO );
  if ( IsThereAFieldFor ( ifd, SamplesPerPixel ) == NO ) return ( NO );

  /* BitsPerSample must have 8, 8, 8 as values */
  if ( ( field = GetFieldStructure ( ifd, BitsPerSample ) )
       == NULL ) return ( NO );
  if ( field->Count != 3 ) return ( NO );
  else { 
    if ( field->Value.UShortArray[0] != 8 ) return ( NO );
    if ( field->Value.UShortArray[1] != 8 ) return ( NO );
    if ( field->Value.UShortArray[2] != 8 ) return ( NO );
  }

  /* Compression must have NoCompression or PackBits as a value */
  if ( ( field = GetFieldStructure ( ifd, Compression ) )
       == NULL ) return ( NO );
  if ( ( field->Value.UShort != NoCompression ) &&
       ( field->Value.UShort != PackBits ) ) return ( NO );

  /* PhotometricInterpretation must have 2 as a value */
  if ( ( field = GetFieldStructure ( ifd, PhotometricInterpretation ) )
       == NULL ) return ( NO );
  if ( field->Value.UShort != 2 ) return ( NO );

  /* SamplesPerPixel must not have a value less than 3 */
  if ( ( field = GetFieldStructure ( ifd, SamplesPerPixel ) )
       == NULL ) return ( NO );
  if ( field->Value.UShort < 3 ) return ( NO );

  /* ResolutionUnit must have 1 or 2 or 3 as a value */
  if ( ( field = GetFieldStructure ( ifd, ResolutionUnit ) )
       == NULL ) return ( NO );
  if ( ( field->Value.UShort != 1 ) &&
       ( field->Value.UShort != 2 ) &&
       ( field->Value.UShort != 3 ) ) return ( NO );

  return ( YES );
}

static int IsImageGrayscale ( struct IFD *ifd ) 
{
  struct TIFF_field *field;

  /* there must be a field for BitsPerSample */
  if ( IsThereAFieldFor ( ifd, BitsPerSample ) == NO ) return ( NO );

  /* BitsPerSample must have 4 or 8 as a value */
  if ( ( field = GetFieldStructure ( ifd, BitsPerSample ) ) 
       == NULL ) return ( NO );
  if ( ( field->Value.UShort != 4 ) &&
       ( field->Value.UShort != 8 ) ) return ( NO );

  /* Compression must have NoCompression or PackBits as a value */
  if ( ( field = GetFieldStructure ( ifd, Compression ) ) 
       == NULL ) return ( NO );
  if ( ( field->Value.UShort != NoCompression ) &&
       ( field->Value.UShort != PackBits ) ) return ( NO );

  /* PhotometricInterpretation   */
  /* must have 0 or 1 as a value */
  if ( ( field = GetFieldStructure ( ifd, PhotometricInterpretation ) ) 
       == NULL ) return ( NO );
  if ( ( field->Value.UShort != 0 ) &&
       ( field->Value.UShort != 1 ) ) return ( NO );

  /* ResolutionUnit must have 1 or 2 or 3 as a value */
  if ( ( field = GetFieldStructure ( ifd, ResolutionUnit ) ) 
       == NULL ) return ( NO );
  if ( ( field->Value.UShort != 1 ) &&
       ( field->Value.UShort != 2 ) &&
       ( field->Value.UShort != 3 ) ) return ( NO );

  /* four BitsPerSample not supported */
  if ( ( field = GetFieldStructure ( ifd, BitsPerSample ) )
       == NULL ) return ( NO );
  if ( field->Value.UShort == 4 ) {
    fprintf ( stderr, "image in TIFF file is grayscale, but image\n" );
    fprintf ( stderr, "data is stored at 4 bits per sample, which\n" );
    fprintf ( stderr, "is not supported by this reader\n" );
    return ( NO );
  }

  return ( YES );
}

static struct TIFF_field *GetFieldStructure ( struct IFD *ifd,
                                              unsigned short Tag )
{
  unsigned short i;
 
  for ( i = 0; i < ifd->NumberOfFields; i++ )
    if ( ifd->Fields[i].Tag == Tag ) return ( &(ifd->Fields[i]) );

  fprintf ( stderr, "tiff.c:  function GetFieldStructure:\n" );
  fprintf ( stderr, "faied to find field with tag %d\n", Tag );
  return ( NULL );
}

static int CheckForCoreFields ( struct IFD *ifd )
{
  if ( ( WhatAboutCoreField ( ifd, ImageWidth ) == ERROR ) ||
       ( WhatAboutCoreField ( ifd, ImageLength ) == ERROR ) ||
       ( WhatAboutCoreField ( ifd, Compression ) == ERROR ) ||
       ( WhatAboutCoreField ( ifd, PhotometricInterpretation ) == ERROR ) ||
       ( WhatAboutCoreField ( ifd, StripOffsets ) == ERROR ) ||
       ( WhatAboutCoreField ( ifd, RowsPerStrip ) == ERROR ) ||
       ( WhatAboutCoreField ( ifd, StripByteCounts ) == ERROR ) ||
       ( WhatAboutCoreField ( ifd, XResolution ) == ERROR ) ||
       ( WhatAboutCoreField ( ifd, YResolution ) == ERROR ) ||
       ( WhatAboutCoreField ( ifd, ResolutionUnit ) == ERROR ) )
    return ( ERROR );

  return ( NO_ERROR );
}

static int WhatAboutCoreField ( struct IFD *ifd, unsigned short Tag )
{
  if ( IsThereAFieldFor ( ifd, Tag ) == NO ) {
    if ( IsThereADefaultFor ( Tag ) == YES ) {
      if ( AddDefaultField ( ifd, Tag ) == ERROR )
        return ( ERROR );
    }
    else {
      TellUserACoreFieldIsNecessary ( Tag );
      return ( ERROR );
    }
  }

  return ( NO_ERROR );
}

static int AddDefaultField ( struct IFD *ifd, unsigned short Tag )
{
  if ( Tag == ResolutionUnit )
    return ( MakeResolutionUnitField ( ifd ) ); 

  fprintf ( stderr, "tiff.c:  function AddDefaultField:\n" );
  fprintf ( stderr, "can not add default field for tag %d\n", Tag );
  return ( ERROR );
}

static void TellUserACoreFieldIsNecessary ( unsigned short Tag )
{
  fprintf ( stderr, "tiff reader:  function CheckForCoreFields:\n" );
  fprintf ( stderr, "field for tag %d does not appear\n", Tag );
  fprintf ( stderr, "in IFD (and has no default)\n" );
}

static int IsThereADefaultFor ( unsigned short Tag )
{
  if ( Tag == ResolutionUnit ) return ( YES );

  return ( NO );
}

static int IsThereAFieldFor ( struct IFD *ifd, unsigned short Tag )
{
  unsigned short i;

  for ( i = 0; i < ifd->NumberOfFields; i++ )
    if ( ifd->Fields[i].Tag == Tag ) return ( YES );

  return ( NO );
}

static int ReadIFD ( FILE *fp, struct IFD *ifd,
                     struct TIFF_header *header )
{
  unsigned short i, NumFieldsInInputFile;

  /* initialize IFD structure */
  ifd->NumberOfFields = 0;
  ifd->Fields = NULL;

  /* get number of fields in IFD of input file */
  if ( GetNumberOfFieldsInInput ( fp, header->OffsetOfFirstIFD, 
                                  &NumFieldsInInputFile ) 
       == ERROR ) return ( ERROR );

  /* loop through fields in IFD */
  for ( i = 0; i < NumFieldsInInputFile; i++ ) {
    /* set file-position at first byte in field */
    if ( SetFilePositionAtFirstByteOfIthField 
         ( fp, header->OffsetOfFirstIFD, i ) == ERROR )
      return ( ERROR );

    /* if tag is recognized, copy field into a structure */
    if ( CopyFieldIfTagIsRecognized ( fp, ifd ) == ERROR ) 
      return ( ERROR );
  }

  /* check to see if this is the last   */
  /* IFD; if it isn't, inform user that */
  /* the others will be ignored         */
  if ( SeeIfThereAreOtherIFDs ( fp, header->OffsetOfFirstIFD, 
                                NumFieldsInInputFile ) == ERROR ) 
    return ( ERROR );

  return ( NO_ERROR );
}

static int SeeIfThereAreOtherIFDs ( FILE *fp, unsigned long OffsetOfIFD,
                                    unsigned short NumFieldsInInputFile )
{
  unsigned long PositionValue, ZeroOrOffset;

  /* compute offset of first byte after end of last IFD */
  PositionValue = OffsetOfIFD + 2 + 
                  ( (unsigned long) (NumFieldsInInputFile) * 12 );

  /* set file-position at first byte after end of last IFD */
  if ( SetFilePosition ( fp, PositionValue ) == ERROR ) {
    fprintf ( stderr, "error going to first byte " );
    fprintf ( stderr, "after end of IFD\n" );
    return ( ERROR );
  }

  /* get 4 bytes */
  if ( GetUnsignedLong ( fp, &ZeroOrOffset ) == ERROR ) {
    fprintf ( stderr, "error getting the four bytes " );
    fprintf ( stderr, "after the end of IFD\n" );
    return ( ERROR );
  }

  /* either this number is zero or   */
  /* it is an offset to another IFD; */
  /* if so, warn user that the other */
  /* IFD(s) will be ignored          */
  if ( ZeroOrOffset != 0 ) {
    fprintf ( stderr, "WARNING:  this TIFF file " );
    fprintf ( stderr, "contains multiple images %ld: %ld;\n", PositionValue, ZeroOrOffset );
    fprintf ( stderr, "all but the first image will be " );
    fprintf ( stderr, "ignored by the TIFF reader\n" );
  }
  
  return ( NO_ERROR );
}

static int SetFilePositionAtFirstByteOfIthField ( FILE *fp, 
                                                  unsigned long OffsetOfIFD,
                                                  unsigned short i )
{
  unsigned long PositionValue;

  /* compute offset of first byte in field */
  PositionValue = OffsetOfIFD + 2 + ( (unsigned long) (i) * 12 );

  /* set file-position at first byte in field */
  if ( SetFilePosition ( fp, PositionValue ) == ERROR ) {
    fprintf ( stderr, "error going to first byte " );
    fprintf ( stderr, "of %d-th field in IFD\n", i );
    return ( ERROR );
  }

  return ( NO_ERROR );
}

static int CopyFieldIfTagIsRecognized ( FILE *fp, struct IFD *ifd ) 
{
  unsigned short TempTag, TempType;
  
  /* get field tag, store in temporary space */
  if ( GetUnsignedShort ( fp, &TempTag ) == ERROR ) {
    fprintf ( stderr, "error reading tag\n" );
    return ( ERROR );
  }

  /* determine whether the tag is recognized; if it */
  /* is not, then the whole field should be ignored */
  if ( IsTagRecognized ( TempTag ) == YES ) {

    /* get value type, store in temporary space */
    if ( GetUnsignedShort ( fp, &TempType ) == ERROR ) {
      fprintf ( stderr, "error reading type of field " );
      fprintf ( stderr, "with tag %d\n", TempTag );
      return ( ERROR );
    }

    /* determine whether the type is recognized; if it */
    /* is not, then the whole field should be ignored  */
    if ( IsTypeRecognized ( TempType ) == YES ) {

      /* determine whether this type is expected for a field */
      /* with this tag; if it is not, then the whole field   */
      /* should be ignored                                   */
      if ( IsTypeExpectedWithTag ( TempTag, TempType ) == YES ) {

        /* increment number of recognized fields */
        ifd->NumberOfFields++;

        /* allocate new field structure, and copy */ 
        /* all field info into the new structure  */
        if ( AllocateAndCopyField ( fp, ifd, TempTag, TempType )
             == ERROR ) return ( ERROR );

      } /* if type is expected with the tag */

    } /* if type is recognized */

  } /* if tag is recognized */

  return ( NO_ERROR );
}

static int AllocateAndCopyField ( FILE *fp, struct IFD *ifd, 
                                  unsigned short Tag, unsigned short Type )
{
  /* allocate a new field structure */
  if ( AllocateNewField ( ifd ) == ERROR ) return ( ERROR );

  /* copy all field info into the new structure */
  if ( CopyField ( fp, &(ifd->Fields[ifd->NumberOfFields - 1]),
                   Tag, Type ) == ERROR ) {
    fprintf ( stderr, "error reading field entry ");
    fprintf ( stderr, "for tag number %d\n", Tag );
    return ( ERROR );
  }

  return ( NO_ERROR );
}

static int IsTypeExpectedWithTag ( unsigned short Tag, unsigned short Type )
{
  if ( Tag == ImageWidth ) {
    if ( Type == SHORT ) return ( YES );
    if ( Type == LONG ) return ( YES );
  }

  if ( Tag == ImageLength ) {
    if ( Type == SHORT ) return ( YES );
    if ( Type == LONG ) return ( YES );
  }

  if ( Tag == BitsPerSample )
    if ( Type == SHORT ) return ( YES );

  if ( Tag == Compression )
    if ( Type == SHORT ) return ( YES );

  if ( Tag == PhotometricInterpretation )
    if ( Type == SHORT ) return ( YES );

  if ( Tag == StripOffsets ) {
    if ( Type == SHORT ) return ( YES );
    if ( Type == LONG ) return ( YES );
  }

  if ( Tag == SamplesPerPixel )
    if ( Type == SHORT ) return ( YES );

  if ( Tag == RowsPerStrip ) {
    if ( Type == SHORT ) return ( YES );
    if ( Type == LONG ) return ( YES );
  }

  if ( Tag == StripByteCounts ) {
    if ( Type == SHORT ) return ( YES );
    if ( Type == LONG ) return ( YES );
  }

  if ( Tag == XResolution )
    if ( Type == RATIONAL ) return ( YES );

  if ( Tag == YResolution )
    if ( Type == RATIONAL ) return ( YES );

  if ( Tag == ResolutionUnit )
    if ( Type == SHORT ) return ( YES );

  if ( Tag == ColorMap )
    if ( Type == SHORT ) return ( YES );

  WrongValueType ( "IsTypeExpectedWithTag", Tag, Type );
  fprintf ( stderr, "field will be ignored\n" );
  return ( NO );
}

static int IsTypeRecognized ( unsigned short Type )
{
  if ( Type == BYTE ) return ( YES );
  if ( Type == SHORT ) return ( YES );
  if ( Type == LONG ) return ( YES );
  if ( Type == RATIONAL ) return ( YES );

  return ( NO );
}

static int CopyField ( FILE *fp, struct TIFF_field *field, 
                       unsigned short Tag, unsigned short Type )
{
  /* put in tag and type */
  field->Tag = Tag; 
  field->Type = Type; 

  /* get number of data values for the field */
  if ( CopyCount ( fp, field ) == ERROR ) {
    fprintf ( stderr, "error reading field count\n" ); 
    return ( ERROR );
  }

  /* get data value(s) */
  if ( CopyValue ( fp, field ) == ERROR ) {
    fprintf ( stderr, "error reading field data value(s)\n" ); 
    return ( ERROR );
  }

  return ( NO_ERROR );
}

static int GetSizeOfType ( struct TIFF_field *field )
{
  if ( field->Type == BYTE ) {
    field->SizeOfType = 1;
    return ( NO_ERROR );
  }
  if ( field->Type == SHORT ) {
    field->SizeOfType = 2; 
    return ( NO_ERROR );
  }
  if ( field->Type == LONG ) {
    field->SizeOfType = 4;
    return ( NO_ERROR );
  }
  if ( field->Type == RATIONAL ) {
    field->SizeOfType = 8;
    return ( NO_ERROR );
  }

  fprintf ( stderr, "tiff.c:  function GetSizeOfType:\n" );
  fprintf ( stderr, "field type %d is not supported\n", 
            field->Type );
  return ( ERROR );
}

static int CopyValue ( FILE *fp, struct TIFF_field *field )
{
  unsigned long Offset;

  /* get size of type */
  if ( GetSizeOfType ( field ) == ERROR ) return ( ERROR );

  /* if value will not fit inside four bytes */
  if ( 4 < field->Count * field->SizeOfType ) {

    /* get offset of location of data value(s) */
    if ( GetUnsignedLong ( fp, &(Offset) ) == ERROR ) {
      fprintf ( stderr, "error reading offset " );
      fprintf ( stderr, "for tag number %d\n", field->Tag );
      return ( ERROR );
    }

    /* set file position where value(s) should be obtained */
    if ( SetFilePosition ( fp, Offset ) == ERROR ) {
      fprintf ( stderr,"error finding data value " );
      fprintf ( stderr, "for tag number %d\n", field->Tag );
      return ( ERROR );
    }

  }

  /* either there is only 1 value or there are multiple values */
  if ( field->Count == 1 ) {
    if ( GetSingleValue ( fp, field ) == ERROR ) {
      fprintf ( stderr, "cound not get data value ");
      fprintf ( stderr, "for tag number %d\n", field->Tag);
      return ( ERROR );
    }
  }
  else if ( field->Count > 1 ) {
    if ( AllocateArrayOfValues ( field ) == ERROR ) {
      fprintf ( stderr, "could not allocate values ");
      fprintf ( stderr, "for tag %d\n", field->Tag);
      return ( ERROR );
    }
    if ( GetArrayOfValues ( fp, field ) == ERROR ) {
      fprintf ( stderr, "cound not get data values ");
      fprintf ( stderr, "for tag number %d\n", field->Tag);
      return ( ERROR );
    }
  }
  else {
    fprintf ( stderr, "tiff.c:  function CopyValue:\ncount for " );
    fprintf ( stderr, "field with tag %d is zero\n", field->Tag );
    return ( ERROR );
  }

  return ( NO_ERROR );
}

static int GetArrayOfValues ( FILE *fp, struct TIFF_field *field )
{
  int i;

  if ( field->Type == BYTE ) {
    for ( i = 0; i < field->Count; i++ )
      if ( GetUnsignedChar ( fp, &(field->Value.UCharArray[i]) )
           == ERROR ) return ( ERROR );
    return ( NO_ERROR );
  }
  else if ( field->Type == SHORT ) {
    for ( i = 0; i < field->Count; i++ )
      if ( GetUnsignedShort ( fp, &(field->Value.UShortArray[i]) )
           == ERROR ) return ( ERROR );
    return ( NO_ERROR );
  }
  else if ( field->Type == LONG ) {
    for ( i = 0; i < field->Count; i++ )
      if ( GetUnsignedLong ( fp, &(field->Value.ULongArray[i]) )
           == ERROR ) return ( ERROR );
    return ( NO_ERROR );
  }
  else if ( field->Type == RATIONAL ) {
    for ( i = 0; i < field->Count; i++ )
      if ( GetRational ( fp, &(field->Value.RatArray[i]) )
           == ERROR ) return ( ERROR );
    return ( NO_ERROR );
  }

  fprintf ( stderr, "tiff.c:  function GetArrayOfValues:\n" );
  fprintf ( stderr, "reader not prepared to get values of\n" );
  fprintf ( stderr, "type %d\n", field->Type );
  return ( ERROR );
}

static int GetSingleValue ( FILE *fp, struct TIFF_field *field )
{
  if ( field->Type == BYTE ) {
    if ( GetUnsignedChar ( fp, &(field->Value.UChar) ) == ERROR )
      return ( ERROR );
    return ( NO_ERROR );
  }
  else if ( field->Type == SHORT ) {
    if ( GetUnsignedShort ( fp, &(field->Value.UShort) ) == ERROR )
      return ( ERROR );
    return ( NO_ERROR );
  }
  else if ( field->Type == LONG ) {
    if ( GetUnsignedLong ( fp, &(field->Value.ULong) ) == ERROR )
      return ( ERROR );
    return ( NO_ERROR );
  }
  else if ( field->Type == RATIONAL ) {
    if ( GetRational ( fp, &(field->Value.Rat) ) == ERROR )
      return ( ERROR );
    return ( NO_ERROR );
  }

  fprintf ( stderr, "tiff.c:  function GetSingleValue:\n" );
  fprintf ( stderr, "reader not prepared to get value of\n" ); 
  fprintf ( stderr, "type %d\n", field->Type ); 
  return ( ERROR );
}

static int AllocateArrayOfValues ( struct TIFF_field *field )
{
  if ( field->Type == BYTE )
    if ( ( field->Value.UCharArray = 
           ( unsigned char * ) 
           calloc ( ( size_t ) field->Count, sizeof ( unsigned char ) ) ) 
         == NULL )
      return ( ERROR );

  if ( field->Type == SHORT )
    if ( ( field->Value.UShortArray =
           ( unsigned short * )
           calloc ( ( size_t ) field->Count, sizeof ( unsigned short ) ) ) 
         == NULL )
      return ( ERROR );

  if ( field->Type == LONG )
    if ( ( field->Value.ULongArray =
           ( unsigned long * )
           calloc ( ( size_t ) field->Count, sizeof ( unsigned long ) ) ) 
         == NULL )
      return ( ERROR );

  if ( field->Type == RATIONAL )
    if ( ( field->Value.RatArray =
           ( struct Rational * )
           calloc ( ( size_t ) field->Count, sizeof ( struct Rational ) ) ) 
         == NULL )
      return ( ERROR );

  return ( NO_ERROR );
}

static int CopyCount ( FILE *fp, struct TIFF_field *field )
{
  if ( GetUnsignedLong ( fp, &(field->Count) ) == ERROR ) {
    fprintf ( stderr, "error reading field count " );
    fprintf ( stderr, "for tag number %d\n", field->Tag );
    return ( ERROR );
  }

  /* count should be positive */
  if ( field->Count == 0 ) {
    fprintf ( stderr, "encountered 0 as field count " );
    fprintf ( stderr, "for tag number %d\n", field->Tag );
    return ( ERROR );
  }

  return ( NO_ERROR );
}

static int AllocateNewField ( struct IFD *ifd )
{
  if ( ( ifd->Fields = (struct TIFF_field *)
                       realloc ( (struct TIFF_field *) ifd->Fields,
                                 ( size_t ) ( ifd->NumberOfFields *
                                 sizeof (struct TIFF_field) ) ) )
       == NULL ) {
    fprintf ( stderr, "tiff.c:  function AllocateNewField\n");
    fprintf ( stderr, "couldn't allocate struct for a new field\n");
    return ( ERROR );
  }

  return ( NO_ERROR );
}

static int IsTagRecognized ( unsigned short TempTag )
{
  if ( TempTag == ImageWidth ) return ( YES );
  if ( TempTag == ImageLength ) return ( YES );
  if ( TempTag == BitsPerSample ) return ( YES );
  if ( TempTag == Compression ) return ( YES );
  if ( TempTag == PhotometricInterpretation ) return ( YES );
  if ( TempTag == StripOffsets ) return ( YES );
  if ( TempTag == SamplesPerPixel ) return ( YES );
  if ( TempTag == RowsPerStrip ) return ( YES );
  if ( TempTag == StripByteCounts ) return ( YES );
  if ( TempTag == XResolution ) return ( YES );
  if ( TempTag == YResolution ) return ( YES );
  if ( TempTag == ResolutionUnit ) return ( YES );
  if ( TempTag == ColorMap ) return ( YES );

  return ( NO );
}

static int GetRational ( FILE *fp, struct Rational *Rat )
{
  unsigned long numerator, denumerator;


  if ( 2 != fscanf ( fp, "%4c%4c", &numerator, &denumerator ) ) {
    fprintf ( stderr,"error reading rational number\n" );
    return ( ERROR );
  }

  if (HOST_ORDER == TIFF_ORDER) {
    Rat->Numer = numerator;
    Rat->Denom = denumerator;
  } else {
    Rat->Numer = LongReverse(numerator);
    Rat->Denom = LongReverse(denumerator);
  }

  return ( NO_ERROR );
}

static int GetUnsignedLong ( FILE *fp, unsigned long *UnsignedLong )
{
  unsigned long unsignedlong;

  if ( 1 != fscanf ( fp, "%4c", &unsignedlong ) ) {
    fprintf ( stderr,"error reading unsigned long\n" );
    return ( ERROR );
  }

  if (HOST_ORDER == TIFF_ORDER)
    *UnsignedLong = unsignedlong;
  else
    *UnsignedLong = LongReverse(unsignedlong);


  return ( NO_ERROR );
}

static int GetUnsignedShort ( FILE *fp, unsigned short *UnsignedShort )
{
  unsigned short unsignedshort;

  if ( 1 != fscanf ( fp, "%2c", &unsignedshort ) ) {
    fprintf ( stderr,"error reading unsigned short\n" );
    return ( ERROR );
  }

  if (HOST_ORDER == TIFF_ORDER)
    *UnsignedShort = unsignedshort;
  else
    *UnsignedShort = ShortReverse(unsignedshort);


  return ( NO_ERROR );
}

static int GetUnsignedChar ( FILE *fp, unsigned char *UnsignedChar )
{
  if ( 1 != fscanf ( fp, "%c", UnsignedChar ) ) {
    fprintf ( stderr,"error reading unsigned char\n" );
    return ( ERROR );
  }

  return ( NO_ERROR );
}

static int GetNumberOfFieldsInInput ( FILE *fp, unsigned long OffsetOfIFD,
                                      unsigned short *NumFieldsInInputFile )
{

  /* set file-position at beginning of IFD */
  if ( SetFilePosition ( fp, OffsetOfIFD ) == ERROR ) {
    fprintf ( stderr,"error finding num of fields for IFD\n" );
    return ( ERROR );
  }

  /* read number of fields in IFD */
  if ( GetUnsignedShort ( fp, NumFieldsInInputFile ) == ERROR ) {
    fprintf ( stderr,"error reading number of fields for IFD\n" );
    return ( ERROR );
  }

  return ( NO_ERROR );
}

static int SetFilePosition ( FILE *fp, unsigned long position )
{
  if ( fseek ( fp, ( long int ) position, SEEK_SET ) < 0 ) {
    fprintf ( stderr,"error setting file-position\n" );
    return ( ERROR );
  }

  return ( NO_ERROR );
}

static int ReadHeader ( FILE *fp, struct TIFF_header *header )
{
  /* set file position at beginning of file */
  if ( SetFilePosition ( fp, 0 ) == ERROR ) return ( ERROR );

  /* get byte-order word */
  if ( GetByteOrder ( fp, header ) == ERROR ) return ( ERROR );
  
  /* ensure second word in header is 42 */
  if ( CheckFortyTwo ( fp, header ) == ERROR ) return ( ERROR );

  /* get offset for first IFD */
  if ( GetOffsetOfFirstIFD ( fp, header ) == ERROR ) return ( ERROR );

  return ( NO_ERROR );
}

static int CheckTypeSizes ( void )
{
  if ( ( 1 != sizeof ( unsigned char ) ) ||
       ( 2 != sizeof ( unsigned short ) ) ||
       ( 4 != sizeof ( unsigned long ) ) ) {
    fprintf ( stderr, "machine data sizes are not " );
    fprintf ( stderr, "appropriate for TIFF reader/writer\n" );
    return ( ERROR );
  }

  return ( NO_ERROR );
}

static int GetByteOrder ( FILE *fp, struct TIFF_header *header )
{
  /* get byte sequence */
  if ( GetUnsignedShort ( fp, &(header->ByteOrder) ) == ERROR ) {
    fprintf ( stderr, "error reading image header\n" );
    return ( ERROR );
  }

  /* ensure that sequence denotes either */
  /* big- or little- endian byte order   */
  TIFF_ORDER = header->ByteOrder;
  if ( header->ByteOrder == BigEndian ) return ( NO_ERROR );
  if ( header->ByteOrder == LittleEndian ) return ( NO_ERROR ) ;
 
  return ( ERROR );
}

static int CheckFortyTwo ( FILE *fp, struct TIFF_header *header )
{
  /* get byte sequence */
  if ( GetUnsignedShort ( fp, &(header->FortyTwo) ) == ERROR ) {
    fprintf ( stderr, "error reading image header\n" );
    return ( ERROR );
  }

  /* compare to the number 42 */
  if ( header->FortyTwo == 42 ) return ( NO_ERROR );
 
  return ( ERROR );
}

static int GetOffsetOfFirstIFD ( FILE *fp, struct TIFF_header *header )
{
  /* get offset for first IFD */
  if ( GetUnsignedLong ( fp, &(header->OffsetOfFirstIFD) ) == ERROR ) {
    fprintf ( stderr, "error reading image header\n" );
    return ( ERROR );
  }

  return ( NO_ERROR );
}

static void PrintIFD ( struct IFD *ifd )
{
  int i;
  
  fprintf ( stdout, "\nno. of fields:\t\t%d\n", ifd->NumberOfFields );

  for ( i = 0; i < ( int ) ( ifd->NumberOfFields ); i++ ) 
    PrintField ( &(ifd->Fields[i]) );

  fprintf ( stdout, "\n" );
}

static void PrintField ( struct TIFF_field *field )
{
  int i;

  fprintf ( stdout, "\ntag:\t\t\t%d\n", field->Tag );
  fprintf ( stdout, "type:\t\t\t%d\n", field->Type );
  fprintf ( stdout, "count:\t\t\t%d\n", field->Count );
    
  if ( field->Count == 1 ) {
    if ( field->Type == BYTE )
      fprintf ( stdout, "value:\t\t\t%d\n", field->Value.UChar );

    if ( field->Type == SHORT )
      fprintf ( stdout, "value:\t\t\t%d\n", field->Value.UShort );

    if ( field->Type == LONG )
      fprintf ( stdout, "value:\t\t\t%d\n", field->Value.ULong );

    if ( field->Type == RATIONAL )
      fprintf ( stdout, "value:\t\t\tnumer:  %d; denom:  %d\n", 
                field->Value.Rat.Numer, field->Value.Rat.Denom );
  }

  if ( field->Count > 1 ) 
    for ( i = 0; i < field->Count; i++ ) {
      if ( field->Type == BYTE )
        fprintf ( stdout, "value[%d]:\t\t%d\n", i, 
                  field->Value.UCharArray[i] );

      if ( field->Type == SHORT )
        fprintf ( stdout, "value[%d]:\t\t%d\n", i, 
                  field->Value.UShortArray[i] );

      if ( field->Type == LONG )
        fprintf ( stdout, "value[%d]:\t\t%d\n", i, 
                  field->Value.ULongArray[i] );

      if ( field->Type == RATIONAL )
        fprintf ( stdout, "value[%d]:\t\tnumer:  %d; denom:  %d\n", 
                  i, field->Value.RatArray[i].Numer,
                  field->Value.RatArray[i].Denom );
    }
}

static void *mget_spc(int num,size_t size)
{
        void *pt;

        if( (pt=malloc((size_t)(num*size))) == NULL ) {
                fprintf(stderr, "==> malloc() error\n");
                exit(-1);
                }
        return(pt);
}

static void **get_img(int wd,int ht,size_t size)
{
        int i;
        void  **ppt;
        char   *pt;

        ppt = (void **)mget_spc(ht,sizeof(void *));
        pt = (char *)mget_spc(wd*ht,size);

        for(i=0; i<ht; i++) ppt[i] = pt + i*wd*size;

        return(ppt);
}

static void free_img(void **pt)
{
        free( (void *)pt[0]);
        free( (void *)pt);
}

static int tiffHostType(int dummy)
{
	unsigned long longsequence[1];
	unsigned char *charsequence;

	longsequence[0] = 0x01020304u;
	charsequence = (unsigned char *) longsequence;
	if      (charsequence[0] == 4) return ((int) LittleEndian);
	else if (charsequence[0] == 1) return ((int) BigEndian);
        else {
/*
		fprintf(stderr, "Unknown Host Byte Order\n");
*/
                return(UnknownEndian);
        }
}
